import numpy as np


def pb_sample(pb_obj, samp_rate, len_min=0, len_max=float('inf'), len_step=1, len_adj=True, debug=False):
    """ Generate sample array.

    Unlike PulseBlock class, which makes now assumptions beyond the base class
    attributes, this function introduces one more requirement:
        each Pulse and DfltPulse object in the block must have
        get_value(t_ar) method with the following signature:
            :param t_ar: (numpy.array) array of time points
            :return: (numpy.array) array of samples

    :param pb_obj: (PulseBlock) pulse block to be sampled
    :param samp_rate: (float) sampling rate [Hz]

    Hardware memory waveform-length limitations

    :param len_min: (int) minimal waveform length (used for sanity check and
    auto-padding if len_adj is set to True)

    :param len_max: (int) maximal waveform length (used for sanity check only)

    :param len_step: (int) waveform length granularity (if length must be an
    integer multiple of a given step)

    :param len_adj: (bool) if True, sample array will be padded with default
    values (determined by pb_odj.dflt_dict) to meet len_min and len_step.
    If False and resulting sample array length does not meet these constraints,
    ValueError will be produced. Note that if length exceeds len_max, ValueError
    is always produced.

    :param debug: (bool) if True, time-point array will be included as a last
    element in the returned tuple

    :return: (tuple)
    (
        samp_dict = {'ch_name': sample_array, ...}
        n_pts - number of samples (per channel, number of time points)
        add_pts - number of default points added to meet len_min and len_step
        [t_ar] (if debug is set to True) - numpy float array of time points
    )
    """

    t_step = 1 / samp_rate
    # Number of samples
    n_pts = int(
        (pb_obj.dur - 0.5*t_step) // t_step + 1
    )
    # '- 0.5*t_step' is added to make actual duration of the waveform to be
    # as close to pb_obj.dur as possible.
    #
    # Normally a waveform-generation device quickly updates output value and
    # keeps it constant for 1/samp_rate interval, then updates output to the
    # new value, and so on. That is why it is impossible to obtain arbitrary
    # duration of the actual waveform - the minimal step is 1/samp_rate.
    #
    # Subtraction of half-step from pb_obj.dur ensures that integer division
    # plus one gives the closest possible number of steps.
    #
    # Note that if pb_obj.dur precisely equals to half-integer number of steps,
    # resulting duration will be pb_obj.dur +/- step/2 and sign may vary randomly
    # (consequence of float-comparison uncertainty for nominally equal floats)

    # Number of points added to meet length constraints
    add_pts = 0

    # Sanity checks -----------------------------------------------------------

    # There is a default pulse for each of the channels in p_dict
    if not set(pb_obj.p_dict.keys()) <= set(pb_obj.dflt_dict.keys()):
        raise ValueError(
            'Every channel with non-default pulses must have the default pulse. \n'
            'The following channels contain non-default pulses: \n'
            '   {} \n'
            'The following channels have default pulse: \n'
            '   {}'
            ''.format(
                sorted(list(pb_obj.p_dict.keys())),
                sorted(list(pb_obj.dflt_dict.keys()))
            )
        )

    # Sample array length fits hardware constraints
    # - length step
    if n_pts % len_step != 0:
        if len_adj:
            new_n_pts = int(
                (n_pts // len_step + 1) * len_step
            )
            add_pts = new_n_pts - n_pts
            n_pts = new_n_pts
        else:
            raise ValueError(
                'Calculated number of points {} does not match hardware step {}. \n'
                'To enable auto-appending of default values, set len_adj to True'
                ''.format(n_pts, len_step)
            )
    # - min length
    if not len_min <= n_pts:
        if len_adj:
            add_pts = len_min - n_pts
            n_pts = len_min
        else:
            raise ValueError(
                'Calculated number of points {} is below hardware minimum {}. \n'
                'Try increasing sampling rate or pulse block duration'
                ''.format(n_pts, len_min)
            )
    # - max length
    if not n_pts <= len_max:
        raise ValueError(
            'Calculated number of points {} is above hardware maximum {}. \n'
            'Try decreasing sampling rate or pulse block duration'
            ''.format(n_pts, len_max)
        )

    # Sample pulse block ------------------------------------------------------

    # Generate arrays of T-points
    t_ar = np.linspace(
        start=0,
        stop=t_step * (n_pts - 1),
        num=n_pts
    )

    # Construct sample array for each channel in dflt_dict.keys()
    # and store it in samp_dict
    samp_dict = dict()

    for ch in pb_obj.dflt_dict.keys():

        # Fill the array with default values
        samp_dict[ch] = pb_obj.dflt_dict[ch].get_value(t_ar=t_ar)

        # Iterate through each pulse item and calculate
        # non-default values for corresponding T-points
        if ch in pb_obj.p_dict.keys():

            for p_item in pb_obj.p_dict[ch]:

                # find indexes of pulse edges
                indx_1 = int(p_item.t0 // t_step)
                indx_2 = int((p_item.t0 + p_item.dur) // t_step)

                # calculate new values
                val_ar = p_item.get_value(
                    t_ar=t_ar[indx_1 : indx_2+1]
                )

                # set the values to sample array
                samp_dict[ch][indx_1 : indx_2+1] = val_ar

    if debug:
        return samp_dict, n_pts, add_pts, t_ar
    else:
        return samp_dict, n_pts, add_pts



""" Generic module for launching pylabnet scripts

NOTE: Requires windows (TODO: make OS agnostic)
NOTE: Only operational on a single machine (TODO: make network compatible)
NOTE: the script module(s) must have a launch() method that can handle kwargs
        :param logger: instance of the LogClient object for logging
        :param logport: (int) port of the log server
        :param clients: dictionary containing all relevant clients in the form e.g.
            {client_1_module.__name__: client_1_instance,
            client_2_module.__name__: client_2_instance, ...}
        :param guis: dictionary containing all relevant GUI clients in the form:
            {ui_filename_1: ui_client_1, ui_filename_2: ui_client_2, ...}
        :param params: arbitrary object containing all script parameters
NOTE: the server module(s) must have a launch() method that can handle kwargs
        :param logger: instance of the LogClient object for loggin
        :param port: port number to use for opening a server

Implements the Launcher class which is used to launch desired pylabnet script(s).
If multiple scripts are provided, they are launched sequentially in this process.
Before launching the script(s), it will check the running GUI's and servers connected
to the main LogServer, and try to instantiate any required servers and GUIs in
separate processes using the pylabnet_gui.py and pylabnet_server.py scripts.

This is meant to be executed by "double-click" action in the log_display GUI, but
can in principle be invoked directly from the command line with appropriate
arguments containing information about the currently running LogServer and all of
its active clients. See log_display.py, Controller._clicked() method for details

Example useage of this class in order to make a script for launching an application
script "monitor_counts.py" that requires a server to be running (provided in
pylabnet.hardware.counter.swabian_instruments.cnt_monitor.py), as well as a GUI
with the template 'count_monitor.ui'. The following .py file should be placed in the
pylabnet.launchers directory

    from .launcher import Launcher
    from pylabnet.hardware.counter.swabian_instruments import cnt_monitor
    from pylabnet.scripts.counter import monitor_counts

    def main():

        launcher = Launcher(
            script=[monitor_counts],
            server_req=[cnt_monitor],
            gui_req=['count_monitor'],
            params=[None]
        )
        launcher.launch()


    if __name__ == '__main__':
        main()

"""

import time
import subprocess
import numpy as np
import sys
import os
import socket
from pylabnet.utils.logging import logger
from pylabnet.utils.helper_methods import parse_args, show_console, hide_console
from pylabnet.network.client_server import external_gui


class Launcher:

    _GUI_LAUNCH_SCRIPT = 'pylabnet_gui.py'
    _SERVER_LAUNCH_SCRIPT = 'pylabnet_server.py'

    def __init__(self, script=None, server_req=None, gui_req=None, auto_connect=True, name=None, params=None):
        """ Instantiates Launcher object

        :param script: script modules to launch. Each module needs to have a launch() method
        :param server_req: list of modules containing necessary servers. The module needs:
            (1) launch() method to instantiate Service and run the server (see pylabnet_server.py for details)
            (2) Client() class, so that we can instantiate a client from this thread and pass it to the script
        :param gui_req: list of gui names to instantiate (names of .ui files, excluding .ui extension)
        :param auto_connect: (bool) whether or not to automatically connect if there is a single instance of the
            required server already running
        :param name: (str) desired name that will appear as the "process" name for the script invoking the
            Launcher object. Can be left blank, and the names of the script module(s) will be used
        :param params: (list) parameters for each script to launch
        """
        self.script = script
        self.server_req = server_req
        self.gui_req = gui_req
        self.auto_connect = auto_connect
        self.params = params
        if name is None:
            if self.script is None:
                self.name = 'Generic Launcher'
            else:
                self.name = ''
                for scr in script:
                    self.name += scr.__name__.split('.')[-1]
                    self.name += '_'
                self.name += 'script'
        else:
            self.name = name

        # Get command line arguments as a dict
        self.args = parse_args()

        try:
            self.log_ip = self.args['logip']
            self.log_port = int(self.args['logport'])
            self.num_clients = int(self.args['numclients'])
            self.debug = int(self.args['debug'])
            self.server_debug = int(self.args['server_debug'])
            self.gui_debug = int(self.args['gui_debug'])
        except IndexError:
            raise

        # Connect to logger.
        self.logger = self._connect_to_logger()

        # Halt execution and wait for debugger connection if debug flag is up.
        if self.debug == 1:
            import ptvsd
            import os
            # 5678 is the default attach port in the VS Code debug configurations
            self.logger.info(f"Waiting for debugger to attach to PID {os.getpid()} (launcher)")
            ptvsd.enable_attach(address=('localhost', 5678))
            ptvsd.wait_for_attach()
            breakpoint()

        # Find all servers with port numbers and store them as a dictionary
        self.connectors = {}
        self._scan_servers()

        # Containers for clients
        self.gui_clients = {}
        self.clients = {}

    def launch(self):
        """ Checks for GUIS/servers, instantiates required, and launches script(s)"""

        try:
            self._launch_guis()
            self._launch_servers()
            self._launch_scripts()
        except Exception as e:
            self.logger.error(e)

    def _connect_to_logger(self):
        """ Connects to the LogServer"""

        log_client = logger.LogClient(host=self.log_ip, port=self.log_port, module_tag=self.name)
        return logger.LogHandler(logger=log_client)

    def _scan_servers(self):
        """ Scans all servers/GUIs connected as clients to the LogServer and adds them to internal attributes"""

        for client_index in range(self.num_clients):

            # Check if there is a port for this client, instantiate connector if so
            port_name = 'port{}'.format(client_index + 1)
            client_name = self.args['client{}'.format(client_index+1)]
            try:
                self.connectors[client_name] = Connector(
                            name=client_name,
                            ip=self.args['ip{}'.format(client_index+1)],
                            port=self.args[port_name]
                        )
            except KeyError:
                pass

            # Check for a ui file as well, if it is a GUI
            ui_name = 'ui{}'.format(client_index + 1)
            try:
                self.connectors[client_name].set_ui(self.args[ui_name])
            except KeyError:
                pass

    def _launch_new_gui(self, gui):
        """ Launches a new GUI and connects to it

        :param gui: (str) name of the .ui file to use as a template
        """

        connected = False
        timeout = 0
        host = socket.gethostbyname(socket.gethostname())

        while not connected and timeout < 1000:
            try:
                gui_port = np.random.randint(1, 9999)
                subprocess.Popen('start /min "{}, {}" /wait "{}" "{}" --logip {} --logport {} --guiport {} --ui {} --debug {}'.format(
                    gui+'_GUI',
                    time.strftime("%Y-%m-%d, %H:%M:%S", time.gmtime()),
                    sys.executable,
                    os.path.join(os.path.dirname(os.path.realpath(__file__)), self._GUI_LAUNCH_SCRIPT),
                    self.log_ip,
                    self.log_port,
                    gui_port,
                    gui,
                    self.gui_debug
                ), shell=True)
                connected = True
            except ConnectionRefusedError:
                self.logger.warn(f'Failed to start {gui} GUI server on {host} with port {gui_port}')
                timeout += 1
                time.sleep(0.01)
        if timeout == 1000:
            self.logger.error(f'Failed to start {gui} GUI server on {host}')
            raise ConnectionRefusedError()

        # Connect to GUI, store client. Try several times, since it may take some time to actually launch the server
        connected = False
        timeout = 0
        while not connected and timeout < 10:
            try:
                self.gui_clients[gui] = external_gui.Client(host=host, port=gui_port)
                connected = True
            except ConnectionRefusedError:
                timeout += 1
                time.sleep(1)

        # If we could connect after roughly 10 seconds, something is wrong and we should raise an error
        if timeout == 10:
            self.logger.error(f'Failed to connect client to newly instantiated {gui} server at \nIP: {host}'
                              f'\nPort: {gui_port}')
            raise ConnectionRefusedError()

    def _connect_to_gui(self, gui, host, port):
        """ Connects to a GUI server with host and port details

        :param gui: (str) name of .ui file to use
        :param host: (str) IP address of server to connect to
        :param port: (int) port number of server to connect to
        """

        self.logger.info('Trying to connect to active GUI Server\nHost: {}\nPort: {}'.format(host, port))
        try:
            self.gui_clients[gui] = external_gui.Client(host=host, port=port)
        except ConnectionRefusedError:
            self.logger.warn('Failed to connect. Instantiating new GUI instead')
            self._launch_new_gui(gui)

    def _launch_guis(self):
        """ Searches through active GUIs to find and connect to/launch relevant ones """

        if self.gui_req is not None:
            for gui in self.gui_req:
                matches = []
                for connector in self.connectors.values():

                    # If we have a match, add it
                    if gui == connector.ui:
                        matches.append(connector)
                num_matches = len(matches)

                # If there are no matches, launch and connect to the GUI manually
                if num_matches == 0:
                    self.logger.info('No active GUIs matching {} were found. Instantiating a new GUI'.format(gui))
                    self._launch_new_gui(gui)

                # If there's 1 match, and we can connect automatically, try that
                elif num_matches == 1 and self.auto_connect:
                    host, port = matches[0].ip, matches[0].port
                    self._connect_to_gui(gui, host, port)

                # If there are multiple matches, force the user to choose in the launched console
                else:
                    msg_str = 'Found relevant GUI(s) already running.\n'
                    self.logger.info(msg_str)
                    show_console()
                    print(msg_str)
                    for index, match in enumerate(matches):
                        msg_str = ('------------------------------------------\n'
                                   + '                    ({})                   \n'.format(index + 1)
                                   + match.summarize())
                        print(msg_str)
                        self.logger.info(msg_str)
                    print('------------------------------------------\n\n'
                          'Which GUI would you like to connect to?\n'
                          'Please enter a choice from {} to {}.'.format(1, len(matches)))
                    use_index = int(input('Entering any other value will launch a new GUI.\n\n>> '))
                    self.logger.info(f'User chose ({use_index})')

                    # If the user's choice falls within a relevant GUI, attempt to connect.
                    try:
                        host, port = matches[use_index-1].ip, matches[use_index-1].port
                        self._connect_to_gui(gui, host, port)

                    # If the user's choice did not exist, just launch a new GUI
                    except IndexError:
                        self.logger.info('Launching new GUI')
                        self._launch_new_gui(gui)
                    hide_console()

    def _launch_new_server(self, module):
        """ Launches a new server

        :param module: (obj) reference to the module which can invoke the relevant server via module.launch()
        """

        connected = False
        timeout = 0
        host = socket.gethostbyname(socket.gethostname())

        while not connected and timeout < 1000:
            try:
                server_port = np.random.randint(1, 9999)
                server = module.__name__.split('.')[-1]

                cmd = 'start /min "{}, {}" /wait "{}" "{}" --logip {} --logport {} --serverport {} --server {} --debug {}'.format(
                    server+"_server",
                    time.strftime("%Y-%m-%d, %H:%M:%S", time.gmtime()),
                    sys.executable,
                    os.path.join(os.path.dirname(os.path.realpath(__file__)), self._SERVER_LAUNCH_SCRIPT),
                    self.log_ip,
                    self.log_port,
                    server_port,
                    server,
                    self.server_debug
                )

                subprocess.Popen(cmd, shell=True)
                connected = True
            except ConnectionRefusedError:
                self.logger.warn(f'Failed to start {server} server on {host} with port {server_port}')
                timeout += 1
                time.sleep(0.01)
            if timeout == 1000:
                self.logger.error(f'Failed to start {server} server on {host}')
                raise ConnectionRefusedError()

        # Connect to server, store client. Try several times, since it may take some time to actually launch the server

        connected = False
        timeout = 0
        while not connected and timeout < 10:
            try:
                self.clients[server] = module.Client(host=host, port=server_port)
                connected = True
            except ConnectionRefusedError:
                timeout += 1

        # If we could connect after roughly 10 seconds, something is wrong and we should raise an error
        if timeout == 10:
            self.logger.error(f'Failed to connect client to newly instantiated {server} server at \nIP: {host}'
                              f'\nPort: {server_port}')
            raise ConnectionRefusedError()

    def _connect_to_server(self, module, host, port):
        """ Connects to a server and stores the client as an attribute, to be used in the main script(s)

        :param module: (object) module from which client can be instantiated using module.Client()
        :param host: (str) IP address of host
        :param port: (int) port number of host
        """

        server = module.__name__.split('.')[-1]
        self.logger.info('Trying to connect to active {} server\nHost: {}\nPort: {}'.format(server, host, port))
        try:
            self.clients[server] = module.Client(host=host, port=port)
        except ConnectionRefusedError:
            self.logger.warn('Failed to connect. Instantiating new server instead')
            self._launch_new_server(module)

    def _launch_servers(self):
        """ Searches through active servers and connects/launches them """

        for module in self.server_req:
            matches = []
            for connector in self.connectors.values():

                # If we find a matching server, add it
                if module.__name__.split('.')[-1]+'_server' == connector.name:
                    matches.append(connector)
            num_matches = len(matches)

            # If there are no matches, launch and connect to the server manually
            if num_matches == 0:
                self.logger.info(f'No active servers matching {module.__name__.split(".")[-1]}'
                                 ' were found. Instantiating a new server')
                self._launch_new_server(module)

            # If there is exactly 1 match, try to connect automatically
            elif num_matches == 1 and self.auto_connect:
                self._connect_to_server(module, host=matches[0].ip, port=matches[0].port)

            # If there are multiple matches, force the user to choose in the launched console
            else:
                show_console()
                msg_str = 'Found relevant server(s) already running.\n'
                self.logger.info(msg_str)
                print(msg_str)
                for index, match in enumerate(matches):
                    msg_str = ('------------------------------------------\n'
                               + '                    ({})                   \n'.format(index + 1)
                               + match.summarize())
                    print(msg_str)
                    self.logger.info(msg_str)
                print('------------------------------------------\n\n'
                      'Which server would you like to connect to?\n'
                      'Please enter a choice from {} to {}.'.format(1, len(matches)))
                use_index = int(input('Entering any other value will launch a new server.\n\n>> '))
                self.logger.info(f'User chose ({use_index})')

                # If the user's choice falls within a relevant GUI, attempt to connect.
                try:
                    host, port = matches[use_index - 1].ip, matches[use_index - 1].port
                    self._connect_to_server(module, host, port)

                # If the user's choice did not exist, just launch a new GUI
                except IndexError:
                    self.logger.info('Launching new server')
                    self._launch_new_server(module)
                hide_console()

    def _launch_scripts(self):
        """ Launch the scripts to be run sequentially in this thread """

        for index, script in enumerate(self.script):

            script.launch(
                logger=self.logger,
                loghost=self.log_ip,
                clients=self.clients,
                guis=self.gui_clients,
                logport=self.log_port,
                params=self.params[index]
            )


class Connector:
    """ Generic container for information about current clients to the LogServer"""

    def __init__(self, name=None, ip=None, port=None):
        """ Instantiates connector

        :param name: (str, optional) name of the client
        :param ip: (str, optional) IP address of the client
        :param port: (str, optional) port number of the client
        """

        self.name = name
        self.ip = ip
        self.port = port
        self.ui = None

    def set_name(self, name):
        """ Sets the name of the connector

        :param name: (str) name of connector
        """
        self.name = name

    def set_ip(self, ip):
        """ Sets the IP address of the connector

        :param ip: (str) IP Address of connector
        """
        self.ip = ip

    def set_port(self, port):
        """ Sets the port number

        :param port: (str) port number
        """
        self.port = int(port)

    def set_ui(self, ui):
        """ Sets the ui filename

        :param ui: (str) ui filename (.ui attachment optional)
        """
        self.ui = ui

    def summarize(self):
        """ Summarizes connector properties. Useful for debugging/logging purposes

        :return: (str) summary of all properties
        """
        return 'Name: {}\nIP: {}\nPort: {}\nUI: {}'.format(self.name, self.ip, self.port, self.ui)




from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase

class Service(ServiceBase):

    def exposed_up(self):
        return self._module.up()

    def exposed_down(self):
        return self._module.down()

    def exposed_get_name(self):
        return self._module.get_name()


class Client(ClientBase):

    def up(self):
        return self._service.exposed_up()

    def down(self):
        return self._service.exposed_down()

    def get_name(self):
        return self._service.exposed_get_name()


""" Generic script for monitoring counts from a counter """

import numpy as np
import time
import socket
from pylabnet.gui.pyqt.gui_handler import GUIHandler
from pylabnet.utils.logging.logger import LogClient
from pylabnet.scripts.pause_script import PauseService
from pylabnet.network.core.generic_server import GenericServer
from pylabnet.utils.helper_methods import unpack_launcher


# Static methods

def generate_widgets():
    """Static method to return systematically named gui widgets for 4ch wavemeter monitor"""

    graphs, legends, numbers = [], [], []
    for i in range(2):
        graphs.append('graph_widget_' + str(i + 1))
        legends.append('legend_widget_' + str(i + 1))
        numbers.append('number_label_' + str(i + 1))
    for i in range(2, 8):
        numbers.append('number_label_' + str(i + 1))
    return graphs, legends, numbers


class CountMonitor:

    # Generate all widget instances for the .ui to use
    _plot_widgets, _legend_widgets, _number_widgets = generate_widgets()

    def __init__(self, ctr_client, gui_client, logger_client):
        """ Constructor for CountMonitor script

        :param ctr_client: (optional) instance of hardware client for counter
        :param gui_client: (optional) instance of client of desired output GUI
        :param logger_client: (obj) instance of logger client.
        """

        self._ctr = ctr_client
        self._bin_width = None
        self._n_bins = None
        self._ch_list = None
        self._plot_list = None  # List of channels to assign to each plot (e.g. [[1,2], [3,4]])
        self._plots_assigned = []  # List of plots on the GUI that have been assigned

        # Instanciate gui handler
        self.gui_handler = GUIHandler(gui_client, logger_client)

    def set_hardware(self, ctr):
        """ Sets hardware client for this script

        :param ctr: instance of count monitor hardware client
        """

        # Initialize counter instance
        self._ctr = ctr

    def set_params(self, bin_width=1e9, n_bins=1e4, ch_list=[1], plot_list=None):
        """ Sets counter parameters
        
        :param bin_width: bin width in ps
        :param n_bins: number of bins to display on graph
        :param ch_list: (list) channels to record
        :param plot_list: list of channels to assign to each plot (e.g. [[1,2], [3,4]])
        """

        # Save params to internal variables
        self._bin_width = int(bin_width)
        self._n_bins = int(n_bins)
        self._ch_list = ch_list
        self._plot_list = plot_list

        # Configure counting channels
        self._ctr.set_channels(ch_list=ch_list)

    def run(self):
        """ Runs the counter from scratch"""

        try:

            # Start the counter with desired parameters
            self._initialize_display()

            # Give time to initialize
            time.sleep(0.05)
            self._is_running = True
            self._ctr.start_counting(bin_width=self._bin_width, n_bins=self._n_bins)

            # Continuously update data until paused
            while self._is_running:
                self._update_output()

        except Exception as exc_obj:
            self._is_running = False
            raise exc_obj

    def pause(self):
        """ Pauses the counter"""

        self._is_running = False

    def resume(self):
        """ Resumes the counter.
        
        To be used to resume after the counter has been paused.
        """

        try:
            self._is_running = True

            # Clear counter and resume plotting
            self._ctr.clear_counter()
            while self._is_running:
                self._update_output()

        except Exception as exc_obj:
            self._is_running = False
            raise exc_obj

    # Technical methods

    def _initialize_display(self):
        """ Initializes the display (configures all plots) """

        plot_index = 0
        for channel in self._ch_list:

            # Figure out which plot to assign to
            if self._plot_list is not None:
                for index, channel_set in enumerate(self._plot_list):
                    if channel in channel_set:
                        plot_index = index
                        break

            # If we have not assigned this plot yet, assign it
            if plot_index not in self._plots_assigned:
                self.gui_handler.assign_plot(
                    plot_widget=self._plot_widgets[plot_index],
                    plot_label='Counter Monitor {}'.format(plot_index + 1),
                    legend_widget=self._legend_widgets[plot_index]
                )
                self._plots_assigned.append(plot_index)

            # Now assign this curve
            self.gui_handler.assign_curve(
                plot_label='Counter Monitor {}'.format(plot_index + 1),
                curve_label='Channel {}'.format(channel),
                error=True
            )

            # Assign scalar
            self.gui_handler.assign_label(
                label_widget=self._number_widgets[channel - 1],
                label_label='Channel {}'.format(channel)
            )

    def _update_output(self):
        """ Updates the output to all current values"""

        # Update all active channels
        # x_axis = self._ctr.get_x_axis()/1e12
        counts = self._ctr.get_counts()
        counts_per_sec = counts*(1e12/self._bin_width)
        noise = np.sqrt(counts)*(1e12/self._bin_width)
        plot_index = 0

        for index, count_array in enumerate(counts_per_sec):

            # Figure out which plot to assign to
            channel = self._ch_list[index]
            if self._plot_list is not None:
                for index_plot, channel_set in enumerate(self._plot_list):
                    if channel in channel_set:
                        plot_index = index_plot
                        break

            # Update GUI data

            self.gui_handler.set_curve_data(
                data=count_array,
                error=noise[index],
                plot_label='Counter Monitor {}'.format(plot_index + 1),
                curve_label='Channel {}'.format(channel)
            )
            self.gui_handler.set_label(
                text='{:.4e}'.format(count_array[-1]),
                label_label='Channel {}'.format(channel)
            )


def launch(**kwargs):
    """ Launches the count monitor script """

    logger, loghost, logport, clients, guis, params = unpack_launcher(**kwargs)

    # Instantiate CountMonitor
    try:
        monitor = CountMonitor(
            ctr_client=clients['si_tt_cnt_monitor'], gui_client=guis['count_monitor'], logger_client=logger
        )
    except KeyError:
        print('Please make sure the module names for required servers and GUIS are correct.')
        time.sleep(15)
        raise

    # Instantiate Pause server
    try:
        pause_logger = LogClient(
            host=loghost,
            port=logport,
            module_tag='count_monitor_pause_server'
        )
    except ConnectionRefusedError:
        logger.warn('Could not connect Count Monitor Pause server to logger')

    pause_service = PauseService()
    pause_service.assign_module(module=monitor)
    pause_service.assign_logger(logger=pause_logger)

    timeout = 0
    while timeout < 1000:
        try:
            port = np.random.randint(1, 9999)
            pause_server = GenericServer(
                host=socket.gethostbyname(socket.gethostname()),
                port=port,
                service=pause_service)
            pause_logger.update_data(data=dict(port=port))
            timeout = 9999
        except ConnectionRefusedError:
            logger.warn(f'Failed to instantiate Count Monitor Pause server at port {port}')
            timeout += 1
    pause_server.start()

    # Set parameters
    if params is None:
        params = dict(bin_width=2e10, n_bins=1e3, ch_list=[1, 2], plot_list=[[1], [2]])
    monitor.set_params(**params)

    # Run
    monitor.run()



import abc


class GatedCtrInterface(abc.ABC):

    @abc.abstractmethod
    def activate_interface(self):
        pass

    @abc.abstractmethod
    def init_ctr(self, bin_number, gate_type):
        """

        :param bin_number:
        :param gate_type: 'RR' and 'RF'
        :return:
        """
        pass

    @abc.abstractmethod
    def close_ctr(self):
        pass

    @abc.abstractmethod
    def start_counting(self):
        pass

    @abc.abstractmethod
    def terminate_counting(self):
        pass

    @abc.abstractmethod
    def get_status(self):
        pass

    @abc.abstractmethod
    def get_count_ar(self, timeout=-1):
        pass


class CtrError(Exception):
    pass



import rpyc
import traceback
import time
import socket
import logging
import sys
import pickle
from pylabnet.utils.helper_methods import get_dated_subdirectory_filepath


class LogHandler:
    """Protection wrapper for logger instance.

    If a module needs to log, always use this class for self.log.

    This class wraps actual logger instance into a protective layer,
    such that no errors of the logger can break the host module.
    This is necessary to avoid reducing module stability due to
    limited stability of the logger (which is often the case).

    Examples of possible errors which a logger can produce:
     - connection lost
     - no logger instance was given

    Such exceptions can break the host module. Since logging is not
    necessary for module operation, it is better to ignore them and
    do not disturb the module.
    """

    def __init__(self, logger=None):
        self._logger = None
        self.set_logger(logger=logger)

    def set_logger(self, logger):
        self._logger = logger

    def debug(self, msg_str):
        try:
            return self._logger.debug(msg_str=msg_str)
        except:
            return -1

    def info(self, msg_str):
        try:
            return self._logger.info(msg_str=msg_str)
        except:
            return -1

    def warn(self, msg_str):
        try:
            return self._logger.warn(msg_str=msg_str)
        except:
            return -1

    def error(self, msg_str):
        try:
            self._logger.error(msg_str=msg_str)
        except:
            return -1

        # All errors will raise exceptions at the LogHandler level.
        raise Exception(msg_str)

    def exception(self, msg_str):
        try:
            return self._logger.exception(msg_str=msg_str)
        except:
            return -1

    def critical(self, msg_str):
        try:
            return self._logger.critical(msg_str=msg_str)
        except:
            return -1


class ILog:
    pass


class LogClient:

    _level_dict = dict(
        NOLOG=100,
        CRITICAL=50,
        ERROR=40,
        WARN=30,
        INFO=20,
        DEBUG=10
    )

    def __init__(self, host, port, module_tag='', server_port=None, ui=None):

        # Declare all internal vars
        self._host = ''
        self._port = 0
        self._connection = None
        self._service = None
        self._level_str = ''
        self._level = 0
        self._module_tag = ''
        self._server_port = server_port  # Identifies a server running in client's thread
        self._ui = ui  # Identifies a relevant .ui file for the client

        # Set module alias to display with log messages
        self._module_tag = module_tag

        # Connect to log server
        self.connect(host=host, port=port)

        # Set module alias to display with log messages
        self._module_tag = module_tag

        # Log test message
        self.info('Started logging')

    def connect(self, host='place_holder', port=-1):

        # Update server address if new values are given
        if host != 'place_holder':
            self._host = host
        if port != -1:
            self._port = port

        # Clean-up old connection if it exists
        if self._connection is not None or self._service is not None:
            try:
                self._connection.close()
            except:
                pass

            self._service = None
            self._connection = None

        # Establish connection if
        if host is None or port is None:
            # 'No logging' was requested.
            #   Do not establish connection to log server
            return 0

        else:
            # Connect to log server
            try:
                self._connection = rpyc.connect(
                    host=self._host,
                    port=self._port,
                    config={'allow_public_attrs': True}
                )
                self._service = self._connection.root

            except Exception as exc_obj:
                self._service = None
                self._connection = None
                self._module_tag = ''

                raise exc_obj

            client_data = dict(
                ip=socket.gethostbyname(socket.gethostname()),
                timestamp=time.strftime("%Y-%m-%d, %H:%M:%S", time.gmtime())
            )
            if self._server_port is not None:
                client_data['port'] = self._server_port
            if self._ui is not None:
                client_data['ui'] = self._ui

            client_data_pickle = pickle.dumps(client_data)
            self._service.add_client_data(self._module_tag, client_data_pickle)
            return 0

    def debug(self, msg_str):
        return self._log_msg(
            msg_str=msg_str,
            level_str='DEBUG'
        )

    def info(self, msg_str):
        return self._log_msg(
            msg_str=msg_str,
            level_str='INFO'
        )

    def warn(self, msg_str):
        return self._log_msg(
            msg_str=msg_str,
            level_str='WARN'
        )

    def error(self, msg_str):
        return self._log_msg(
            msg_str=msg_str,
            level_str='ERROR'
        )

    def exception(self, msg_str):
        # Get traceback string from the last exception
        tb_str = traceback.format_exc()

        # Prepend user-give message
        full_msg_str = msg_str + '\n' + tb_str

        return self.error(msg_str=full_msg_str)

    def critical(self, msg_str):
        return self._log_msg(
            msg_str=msg_str,
            level_str='CRITICAL'
        )

    def update_data(self, data=None):
        """ Updates client data with new information

        :param data: (dict) containing client data
            e.g. {'port': 4444, 'other-data': XYZ}
        """

        if data is not None:
            client_data_pickle = pickle.dumps(data)
            self._service.update_client_data(self._module_tag, client_data_pickle)

    def _log_msg(self, msg_str, level_str):

        # Prepending log message with module name.
        message = ' {0}: {1}'.format(self._module_tag, msg_str)

        # Try sending message to the log server
        try:
            ret_code = self._service.exposed_log_msg(
                msg_str=message,
                level_str=level_str
            )
            return ret_code

        # If connection was lost (EOFError)
        # or was not initialized (AttributeError),
        # try to reconnect and send the message again
        except (EOFError, AttributeError):
            # Try reconnecting.
            #   If an exception is risen at this step,
            #   one cannot fix the problem automatically and
            #   the exception should just be returned to the caller.
            #   That is why no exception catch should be done here.
            print('DEBUG: no connection. Trying to reconnect to log server')
            self.connect()

            # Try sending message again.
            #   If an exception is risen at this step,
            #   one cannot fix problem automatically and
            #   the exception should just be returned to the caller
            #   That is why no exception catch should be done here.
            print('DEBUG: Trying to resend the message')
            ret_code = self._service.exposed_log_msg(
                msg_str=message,
                level_str=level_str
            )
            return ret_code


class LogService(rpyc.Service):

    def __init__(self, form_string=None, console_level=logging.DEBUG, file_level=logging.DEBUG, log_output=False,
                 name=None, dir_path=None):
        """ Instantiate LogService

        If the log_output flag is set to True, a .log file will be generated.

        :form_string: String specifying the output format of the logger. If None, default styling is:
            "%(asctime)s - %(levelname)s - %(message)s"
        :console_level: The minimum message level which appears in the console.
        :file_level: The minimum message level which appears in the log-file.
        :log_output: Boolean indicating if a file output should be generated. If true, the following
            two parameters must be provided:
        :name: Name of the log-file.
        :dir_path: Directory where the log files will be generated.

        Note: A dated subdirectory structure will be automatically generated, such that the
        log output will be logged in the following file:
            'dir_path/YEAR/MONTH/DAY/name.log'
        """

        super().__init__()

        # Note: This is inspired by https://docs.python.org/3/howto/logging-cookbook.html

        # Start instance og logging.logger
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)

        # create console handler and set level to debug
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(console_level)

        formatting_string = '%(asctime)s - %(levelname)s - %(message)s' or form_string

        # create formatter
        formatter = logging.Formatter(formatting_string)

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        self.logger.addHandler(ch)

        # create file handler which logs even debug messages
        if log_output:

            assert name, 'Please provide a name for the logger.'
            assert dir_path, 'Please provide a directory path for the .log file.'

            filename = get_dated_subdirectory_filepath(dir_path, name)
            fh = logging.FileHandler(filename)
            fh.setLevel(file_level)

            # add formatter to fh
            fh.setFormatter(formatter)

            # add fh to logger
            self.logger.addHandler(fh)

        self.client_data = {}
        self.data_updated = []  # Identifies which clients have updated data

    def on_connect(self, conn):
        # code that runs when a connection is created
        # (to init the service, if needed)
        self.logger.info('Client connected')

    def on_disconnect(self, conn):
        # code that runs after the connection has already closed
        # (to finalize the service, if needed)
        self.logger.debug('Client disconnected')

    def exposed_log_msg(self, msg_str, level_str):

        if level_str == 'DEBUG':
            self.logger.debug(msg_str)

        elif level_str == 'INFO':
            self.logger.info(msg_str)

        elif level_str == 'WARN':
            self.logger.warning(msg_str)

        elif level_str == 'ERROR':
            self.logger.error(msg_str)

        return 0

    def add_client_data(self, module_name, module_data_pickle):
        """ Add new client info

        :param module_name: (str) name of the module
        :param module_data_pickle: (pickle) pickled dictionary containing client data.
            e.g. {'ip_address':'0.0.0.0', 'timestamp':'2020-03-04, 12:12:12, 'data':None}
        """

        # Check if this module has already been inserted and modify its name accordingly
        mn = module_name
        module_index = 2
        while module_name in self.client_data:
            module_name = mn + str(module_index)
            module_index += 1

        # Add client data to attribute of service
        self.client_data[module_name] = pickle.loads(module_data_pickle)
        return 0

    def update_client_data(self, module_name, module_data_pickle):
        """ Update client info

        :param module_name: (str) name of module
        :param module_data_pickle: (pickle) pickled dictionary containing client data.
            e.g. {'port': 4444, 'other-data': XYZ}
        """

        try:
            self.client_data[module_name].update(pickle.loads(module_data_pickle))
            self.logger.info('Updated client data for {}'.format(module_name))
            self.data_updated.append(module_name)
        except IndexError:
            self.logger.warning('Tried to update client data for {}, but could not find it in list of clients!'.format(
                module_name
            ))




import abc


class SimplePGenInterface(abc.ABC):

    @abc.abstractmethod
    def activate_interface(self):
        pass

    @abc.abstractmethod
    def write(self, pb_obj, step_adj=True):
        pass

    @abc.abstractmethod
    def set_rep(self, rep_num):
        pass

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    @abc.abstractmethod
    def get_status(self):
        """Get status of the device

        0 - 'Idle'
        1 - 'Running'
        Exception is produced in the case of any error
        (for example, connection to the device is lost)

        :return: (int) status code
                 Exception is produced in the case of error
        """

        pass


class PGenError(Exception):
    pass



""" Launches the staticline GUI test"""

from pylabnet.launchers.launcher import Launcher
from pylabnet.launchers.servers import staticline_nidaqmx
from pylabnet.scripts.staticlines import staticline_gui_generic


def main():

    launcher = Launcher(
        script=[staticline_gui_generic],
        server_req=[staticline_nidaqmx],
        gui_req=['staticline_generic'],
        params=[None]
    )
    launcher.launch()


if __name__ == '__main__':
    main()



import TimeTagger as TT
import time
import copy
import pickle
import numpy as np
from pylabnet.utils.logging.logger import LogHandler
from pylabnet.hardware.interface.gated_ctr import GatedCtrInterface, CtrError
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class Wrap(GatedCtrInterface):

    def __init__(self, tagger, click_ch, gate_ch, logger=None):
        """Instantiate gated counter

        :param tagger: instance of TimeTagger class
        :param click_ch: (int|list of int) clicks on all specified channels
                                    will be summed into one logical channel
        :param gate_ch: (int) positive/negative channel number - count while
                             gate is high/low
        """

        # Log
        self.log = LogHandler(logger=logger)

        # Reference to tagger
        self._tagger = tagger

        # Log device ID information to demonstrate that connection indeed works
        serial = self._tagger.getSerial()
        model = self._tagger.getModel()
        self.log.info(
            'Got reference to Swabian Instruments TimeTagger device \n'
            'Serial number: {0}, Model: {1}'
            ''.format(serial, model)
        )

        # Gated Counter
        # reference to the TT.CountBetweenMarkers measurement instance
        self._ctr = None
        # number of count bins:
        #   length of returned 1D count array, the expected number of gate pulses,
        #   the size of allocated memory buffer.
        # must be given as argument of init_ctr() call
        self._bin_number = 0

        # Channel assignments
        self._click_ch = 0
        self._gate_ch = 0
        # reference to Combiner object
        #   (if _click_ch is a list - then counts on all channels are summed
        #   into virtual channel - self._combiner.getChannel())
        self._combiner = None
        # apply channel assignment
        self.set_ch_assignment(
            click_ch=click_ch,
            gate_ch=gate_ch
        )

        # Module status code
        #  -1 "void"
        #   0 "idle"
        #   1 "in_progress"
        #   2 "finished"
        self._status = -1
        self._set_status(-1)

        # Once __init__() call is complete,
        # the counter is ready to be initialized by the above-lying logic though init_ctr() call

    # ---------------- Interface ---------------------------

    def activate_interface(self):
        return 0

    def init_ctr(self, bin_number, gate_type):

        # Device-specific fix explanation:
        #
        #   CountBetweenMarkers measurement configured for n_value bins
        #   indeed fills-up buffer after n_value gate pules, but call of
        #   self._ctr.ready() still gives False. Only after one additional
        #   gate pulse it gives True, such that self.get_status()
        #   gives 2 and self.get_count_ar() returns:
        #
        #       device always needs an additional pulse to complete
        #       (even for n_values = 1 it needs 2 gate pulses).
        #
        #   Since this is very counter-intuitive and confuses
        #   above-lying logic, a fix is made here:
        #
        #       For given bin_number, CountBetweenMarkers measurement
        #       is instantiated with n_values = (bin_number - 1) such
        #       that it completes after receiving bin_number physical
        #       gate pulses as expected.
        #
        #       As a result, in the returned count_ar
        #       (still of length bin_number, as logic expects), the last
        #       value is just a copy of (bin_number - 1)-st element.
        #
        #   The last physical bin is not actually measured, what can lead
        #   to confusions when bin_number is on the order of one.
        #   The warning below reminds about it:
        if bin_number <= 5:
            self.log.warn(
                'init_ctr(): due to strange behaviour of TT.CountBetweenMarkers '
                'measurement, this driver makes a hack: counter is configured to '
                'measure bin_number-1 pulses and the last element of the returned '
                'count_ar is just a copy of the preceding one. \n'
                'With bin_number={}, only the first {} gate windows will actually be '
                'measured.'
                ''.format(bin_number, bin_number - 1)
            )

        # Close existing counter, if it was initialized before
        if self.get_status() != -1:
            self.close_ctr()

        # Instantiate counter measurement
        try:
            if gate_type == 'RF':
                self._ctr = TT.CountBetweenMarkers(
                    tagger=self._tagger,
                    click_channel=self._click_ch,
                    begin_channel=self._gate_ch,
                    end_channel=-self._gate_ch,
                    n_values=bin_number - 1
                )
            elif gate_type == 'RR':
                self._ctr = TT.CountBetweenMarkers(
                    tagger=self._tagger,
                    click_channel=self._click_ch,
                    begin_channel=self._gate_ch,
                    n_values=bin_number - 1
                )
            else:
                msg_str = 'init_ctr(): unknown gate type "{}" \n' \
                          'Valid types are: \v' \
                          '     "RR" - Raising-Raising \n' \
                          '     "RF" - Raising-Falling'
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

            # set status to "idle"
            self._set_status(0)

            # save bin_number in internal variable
            self._bin_number = bin_number

        # handle NotImplementedError (typical error, produced by TT functions)
        except NotImplementedError:
            # remove reference to the counter measurement
            self._ctr = None
            # set status to "void"
            self._set_status(-1)

            msg_str = 'init_ctr(): instantiation of CountBetweenMarkers measurement failed'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

        # Prepare counter to be started by start_counting()
        # (CountBetweenMarkers measurement starts running immediately after instantiation,
        # so it is necessary to stop it and erase all counts collected between instantiation and stop() call)
        self._ctr.stop()
        self._ctr.clear()

        return 0

    def close_ctr(self):

        # Try to stop and to clear TT.CountBetweenMarkers measurement instance
        try:
            self._ctr.stop()
            self._ctr.clear()
        except:
            pass

        # Remove reference, set status to "void"
        self._ctr = None
        self._set_status(-1)

        return 0

    def start_counting(self):

        current_status = self.get_status()

        # Sanity check: ensure that counter is not "void"
        if current_status == -1:
            msg_str = 'start_counting(): ' \
                      'counter is in "void" state - it ether was not initialized or was closed. \n' \
                      'Initialize it by calling init_ctr()'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

        # Terminate counting if it is already running
        if current_status == 1:
            self.terminate_counting()

        # Try stopping and restarting counter measurement
        try:
            self._ctr.stop()  # does not fail even if the measurement is not running
            self._ctr.clear()
            self._ctr.start()

            # set status to "in_progress"
            self._set_status(1)

            # Wait until the counter is actually ready to count
            time.sleep(0.1)

            return 0

        # handle exception in TT function calls [NotImplementedError]
        except NotImplementedError:
            # Since stop() and clear() methods are very robust,
            # this part is only executed if counter is totally broken.
            # In this case it makes sense to close counter.
            self.close_ctr()

            msg_str = 'start_counting(): call failed. Counter was closed. \n'\
                      'Re-initialize counter by calling init_ctr() again'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

    def terminate_counting(self):

        # Action of this method is non-trivial for "in_progress" state only
        if self.get_status() != 1:
            return 0

        # Try stopping and clearing counter measurement
        try:
            # stop counter, clear count array
            self._ctr.stop()
            self._ctr.clear()

            # set status to "idle"
            self._set_status(0)
            return 0

        # handle exception in TT.stop()/TT.clear()
        except NotImplementedError:
            # Since stop() and clear() methods are very robust,
            # this part is only executed if counter is totally broken.
            # In this case it makes sense to close counter.
            self.close_ctr()

            msg_str = 'terminate_counting(): call failed. Counter was closed. \n' \
                      'Re-initialize it by calling init_ctr()'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

    def get_status(self):

        # Check that counter measurement was initialized and that the connection works
        # by calling isRunning()
        #  -- if self._ctr is None or if connection is broken, call will rise some
        #     exception. In this case "void" status should be set
        #  -- if counter was initialized and connection works, it will return successfully
        #     (True or False, but the result does not matter)
        #     and further choice between "idle", "in_progress", and "finished" should be made
        try:
            self._ctr.isRunning()
        except:
            # set status to "void"
            self._status = -1

        # No handling of "idle" and "finished" status is needed:
        # it will be returned by self._status as is

        # Handle "in_progress" status
        #   This status means that measurement was started before.
        #   Now one needs to check if it is already finished or not.
        #   If measurement is complete, change status to "finished".
        if self._status == 1:
            if self._ctr.ready():

                self._ctr.stop()
                self._status = 2

        return copy.deepcopy(self._status)

    def get_count_ar(self, timeout=-1):

        # If current status is "in_progress",
        # wait for transition to some other state:
        #   "finished" if measurement completes successfully,
        #   "idle" if measurement is terminated,
        #   "void" if counter breaks
        start_time = time.time()
        sleep_time = abs(timeout)/100

        while self.get_status() == 1:
            # stop waiting if timeout elapses
            if time.time()-start_time > timeout >= 0:
                break
            time.sleep(sleep_time)

        # Analyze current status and return correspondingly
        status = self.get_status()

        # return data only in the case of "finished" state
        if status == 2:
            count_array = np.array(
                self._ctr.getData(),
                dtype=np.uint32
            )

            # Fix of the issue with an additional gate pulse needed to complete
            # measurement (see comment in init_ctr() for explanation):
            #   the last element of returned array is just a copy
            #   of the last physically measured bin
            count_array = np.append(count_array, count_array[-1])

            return count_array

        # return empty list for all other states ("in_progress", "idle", and "void")
        else:
            if status == 1:
                self.log.warn(
                    'get_count_ar(): operation timed out, but counter is still running. \n'
                    'Try calling get_count_ar() later or terminate process by terminate_counting().'
                )
            elif status == 0:
                self.log.warn('get_count_ar(): counter is "idle" - nothing to read')
            else:
                msg_str = 'get_count_ar(): counter broke and was deleted \n' \
                          'Re-initialize it by calling init_ctr()'
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

            return []

    # ------------------------------------------------------

    def _set_status(self, new_status):
        """Method to set new status in a clean way.

        This method compares the requested new_status with current status
        and checks if this transition is possible. If transition is possible,
        the change is applied to self._status. Otherwise, no status change
        is applied, -1 is returned, and error message is logged.


        :param new_status: (int) new status value
                            -1 - "void"
                             0 - "idle"
                             1 - "in_progress"
                             2 - "finished"

        :return: (int) operation status code:
                 0 - OK, change was accepted and applied
                -1 - Error, impossible transition was requested,
                     no state change was applied
        """

        # Transition to "void" is always possible
        # by calling close_ctr()
        if new_status == -1:
            self._status = -1
            return 0

        # Transition to "idle" is possible from
        #   "void" by calling init_ctr()
        #   "in_progress" by calling terminate_counting()
        if new_status == 0:
            if self._status == -1 or self._status == 1:
                self._status = 0
                return 0
            else:
                msg_str = '_set_status(): transition to new_status={0} from self._status={1} is impossible. \n'\
                          'Counter status was not changed.'\
                          ''.format(new_status, self._status)
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

        # Transition to "in_progress" is possible from
        #   "idle" by calling start_counting()
        #   "finished" by calling start_counting()
        if new_status == 1:
            if self._status == 0 or self._status == 2:
                self._status = 1
                return 0
            else:
                msg_str = '_set_status(): transition to new_status={0} from self._status={1} is impossible. \n'\
                          'Counter status was not changed.'\
                          ''.format(new_status, self._status)
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

        # Transition to "finished" is only possible from "in_progress"
        # by successful completion of count_array accumulation
        if new_status == 2:
            if self._status == 1:
                self._status = 2
                return 0
            else:
                msg_str = '_set_status(): transition to new_status={0} from self._status={1} is impossible. \n'\
                          'Counter status was not changed.'\
                          ''.format(new_status, self._status)
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

    def get_ch_assignment(self):
        """Returns dictionary containing current channel assignment:
            {
                'click_ch': (int) click_channel_number_including_edge_sign
                'gate_ch': (int) gate_channel_number_including_edge_sign
            }

        :return: dict('click_ch': _, 'gate_ch': _)
        """

        click_ch = copy.deepcopy(self._click_ch)
        gate_ch = copy.deepcopy(self._gate_ch)

        return dict(click_ch=click_ch, gate_ch=gate_ch)

    def set_ch_assignment(self, click_ch=None, gate_ch=None):
        """Sets click channel and and gate channel.

        This method only changes internal variables
        self._click_ch and self._gate_ch.
        To apply the channel update, call  init_ctr() again.


        :param click_ch: (int|list of int) click channel number
                              positive/negative values - rising/falling edge detection
                              if list is given, clicks on all specified channels
                              will be merged into one logic channel

        :param gate_ch: (int) channel number
                             positive/negative - count during high/low gate level

        :return: (dict) actually channel assignment:
                        {
                            'click_channel': (int) click_chnl_num,
                            'gate_channel': (int) gate_chnl_num
                        }
        """

        if click_ch is not None:
            # for convenience bring int type of input to list of int
            if isinstance(click_ch, list):
                click_ch_list = click_ch
            elif isinstance(click_ch, int):
                click_ch_list = [click_ch]
            else:
                # unknown input type
                msg_str = 'set_ch_assignment(click_ch={0}): invalid argument type'\
                          ''.format(click_ch)
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

            # sanity check: all requested channels are available on the device
            all_chs = self.get_all_chs()
            for channel in click_ch_list:
                if channel not in all_chs:
                    msg_str = 'set_ch_assignment(): '\
                              'click_ch={0} - this channel is not available on the device'\
                              ''.format(click_ch)
                    self.log.error(msg_str=msg_str)
                    raise CtrError(msg_str)

            # If several channel numbers were passed, create virtual Combiner channel
            if len(click_ch_list) > 1:
                self._combiner = TT.Combiner(
                    tagger=self._tagger,
                    channels=click_ch_list
                )
                # Obtain int channel number for the virtual channel
                click_ch_list = [self._combiner.getChannel()]

            # Set new value for click channel
            self._click_ch = int(click_ch_list[0])

        if gate_ch is not None:

            # sanity check: channel is available on the device
            if gate_ch not in self.get_all_chs():
                msg_str = 'set_ch_assignment(): '\
                          'gate_ch={0} - this channel is not available on the device'\
                          ''.format(gate_ch)
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

            # Set new value for gate channel
            self._gate_ch = int(gate_ch)

        return self.get_ch_assignment()

    def get_all_chs(self):
        """Returns list of all channels available on the device,
        including edge type sign.

        Positive/negative numbers correspond to detection of rising/falling edges.
        For example:
            1 means 'rising edge on connector 1'
            -1 means 'falling edge on connector 1


        :return: (list of int) list of channel numbers including edge sign.
                Example: [-8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8]
                Empty list is returned in the case of error.
        """

        # Sanity check: check that connection to the device was established
        if self._tagger is None:
            msg_str = 'get_all_chs(): not connected to the device yet'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

        channel_list = list(
            self._tagger.getChannelList(TT.TT_CHANNEL_RISING_AND_FALLING_EDGES)
        )
        return channel_list


class Service(ServiceBase):

    def exposed_activate_interface(self):
        return self._module.activate_interface()

    def exposed_init_ctr(self, bin_number, gate_type):
        return self._module.init_ctr(
            bin_number=bin_number,
            gate_type=gate_type
        )

    def exposed_close_ctr(self):
        return self._module.close_ctr()

    def exposed_start_counting(self):
        return self._module.start_counting()

    def exposed_terminate_counting(self):
        return self._module.terminate_counting()

    def exposed_get_status(self):
        return self._module.get_status()

    def exposed_get_count_ar(self, timeout=-1):
        res = self._module.get_count_ar(timeout=timeout)
        return pickle.dumps(res)


class Client(ClientBase, GatedCtrInterface):

    def activate_interface(self):
        return self._service.exposed_activate_interface()

    def init_ctr(self, bin_number, gate_type):
        return self._service.exposed_init_ctr(
            bin_number=bin_number,
            gate_type=gate_type
        )

    def close_ctr(self):
        return self._service.exposed_close_ctr()

    def start_counting(self):
        return self._service.exposed_start_counting()

    def terminate_counting(self):
        return self._service.exposed_terminate_counting()

    def get_status(self):
        return self._service.exposed_get_status()

    def get_count_ar(self, timeout=-1):
        res_pickle = self._service.exposed_get_count_ar(timeout=timeout)
        return pickle.loads(res_pickle)



""" Client to access qudi's confocal refocus
"""


from pylabnet.network.core.client_base import ClientBase


class RefocusClient(ClientBase):
    def rfcs(self):
        return self._service.exposed_refocus()



import numpy as np
import socket
import sys
from PyQt5 import QtWidgets, uic, QtCore

from pylabnet.network.core.generic_server import GenericServer
from pylabnet.network.client_server.external_gui import Service, Client
from pylabnet.gui.pyqt.external_gui import Window
from pylabnet.utils.helper_methods import show_console, hide_console


def launch(logger=None, port=None, name=None):
    """ Instantiates a default main window + server

    :param logger: (LogClient) instance of LogClient for logging purposes
    :param port: (int) port number for the GUI server
    :param name: (str) name of server to display
    """

    # Create app and instantiate main window
    app = QtWidgets.QApplication(sys.argv)
    show_console()
    ui = input('Please enter the .ui file to use as a template:\n>> ')
    hide_console()
    try:
        main_window = Window(app, gui_template=ui)
    except FileNotFoundError:
        print('Could not find .ui file, '
              'please check that it is in the pylabnet/gui/pyqt/gui_templates directory')
        raise
    gui_service = Service()
    gui_service.assign_module(module=main_window)
    gui_service.assign_logger(logger=logger)

    if port is None:
        port = np.random.randint(1, 9999)
    gui_server = GenericServer(
        service=gui_service,
        host=socket.gethostbyname(socket.gethostname()),
        port=port
    )

    gui_server.start()

    # Update GUI with server-specific details
    main_window.ip_label.setText('IP Address: {}'.format(
        socket.gethostbyname(socket.gethostname())
    ))
    main_window.port_label.setText('Port: {}'.format(port))

    # Run the GUI until the stop button is clicked
    while not main_window.stop_button.isChecked():
        main_window.configure_widgets()
        main_window.update_widgets()
        main_window.force_update()
    sys.exit(app.exec_())



import unicodedata
import os
import time
import re
import sys
import ctypes
import numpy as np
from pylabnet.network.core.generic_server import GenericServer


def str_to_float(in_val):
    """Convert human-readable exponential form to float.

    :param in_val: (str) input string of the following formats:

            'float_number' --> float_number

            'float_number + white_space + exp_prefix + unit_string'
                                 --> float_number * 10**exp_value

        Supported exp prefixes: ['T', 'G', 'M', 'k', '', 'm', 'u', 'n', 'p']

        Warning: format 'just exp_prefix without unit_string' is not
        supported: if only one symbol is given after 'float_number',
        it will be interpreted as unit and exponent will be set to 10**0.

        Examples: '1.2 us'   --> 1.2e-6
                  '-4.5 mV'  --> -4.5e-3
                  '10.1 GHz' --> 1.01e10
                  '1.56 s'   --> 1.56
                  '1.56 m'   --> 1.56 [interpreted as 1.56 meters, not as 1.56e-3]

    :return: (float) extracted value without unit
    """

    if isinstance(in_val, (float, int)):
        return in_val

    # Split string into mantissa and exp_prefix + unit
    item_list = in_val.split()

    # Extract mantissa exp_prefix if included
    mantissa = float(item_list[0])
    # Extract exp_prefix (a single letter) if included
    try:
        exp_prefix_unit = item_list[1]

        if len(exp_prefix_unit) > 1:
            exp_prefix = item_list[1][0]
        else:
            exp_prefix = ''

    except IndexError:
        exp_prefix = ''

    # Convert exp_prefix into exp_value
    if exp_prefix == 'T':
        exp_value = 12
    elif exp_prefix == 'G':
        exp_value = 9
    elif exp_prefix == 'M':
        exp_value = 6
    elif exp_prefix == 'k':
        exp_value = 3
    elif exp_prefix == '':
        exp_value = 0
    elif exp_prefix == 'm':
        exp_value = -3
    elif exp_prefix == 'u':
        exp_value = -6
    elif exp_prefix == 'n':
        exp_value = -9
    elif exp_prefix == 'p':
        exp_value = -12
    else:
        # The case of multi-letter unit without prefix: '1.5 Hz'
        # the first letter 'H' is not an exp prefix
        exp_value = 0

    return mantissa * (10 ** exp_value)


def pwr_to_float(in_val):

    # FIXME: implement

    # if isinstance(in_val, float):
    #     return in_val
    # #
    # # Determine whether the power is given in Volts or dBm
    # #
    # # Split string into mantissa and exp_prefix + unit
    # item_list = in_val.split()
    #
    # # Extract exp_prefix (a single letter) if included
    # try:
    #     exp_prefix_unit = item_list[1]
    #
    #     if len(exp_prefix_unit) > 1:
    #         exp_prefix = item_list[1][0]
    #     else:
    #         exp_prefix = ''
    # except IndexError:
    #     exp_prefix = ''

    return str_to_float(in_val=in_val)


def slugify(value, allow_unicode=False):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.

    From Django 2.2
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value.lower()).strip()
    return re.sub(r'[-\s]+', '-', value)


def get_dated_subdirectory_filepath(directory, filename):
    '''Creates directory structure folder_path/YEAR/MONTH/DAY/filename

    :folder_path: Upper level directory
    :filename: Name of file. Will be slugified.

    Return:
    :filepath: Path to file in newly created structure.
    '''

    # Create subdirectory structure: YEAR/MONTH/DAY
    dated_path = os.path.join(directory, time.strftime('%Y'), time.strftime('%m'), time.strftime('%d'))

    # create folders if they don't exists yet
    os.makedirs(dated_path, exist_ok=True)

    # Define full file path
    filepath = os.path.join(dated_path, f'{slugify(filename)}.log')

    return filepath


def dict_to_str(dic, separate='\n'):
    """ Converts a dictionary to a nicely formatted string

    :param dic: (dict) to convert
    :param separate: (str) string to use to separate dictionary elements
    :return: (str) of dictionary content
    """

    dict_str = ''
    for key, value in dic.items():
        dict_str += '{}: {}{}'.format(key, value, separate)

    return dict_str.rstrip()


def remove_spaces(st):
    """ Removes spaces from a string

    :param st: (str) input string with spaces
    :return: (str) string without any spaces
    """

    return st.replace(' ', '')


def parse_args():
    """ Parses command line arguments into dictionary format, assuming keywords of the form --kw for each argument"""

    arg_index = 1
    arg_dict = {}
    while arg_index < len(sys.argv) - 1:
        arg_dict[sys.argv[arg_index][2:]] = sys.argv[arg_index + 1]
        arg_index += 2

    return arg_dict


def unpack_launcher(**kwargs):
    """ Unpacks the launcher kwargs for easy use in launcher method definition within script modules.
    Copy paste the following implementation at the top of script.launch() method:

    logger, loghost logport, clients, guis, params = unpack_launcher(**kwargs)

    :param kwargs: (dict) contains all keyword arguments required for launching a script from launcher module
        e.g.: dict(logger=log, loghost='localhost', clients=[client1, client2], guis=[gui_client1, gui_client2]),
                   logport=1234, params=experimental_parameters_container)
        Note that experimental parameters should go in "params" and can be customized to contain all other
        script specific stuff

    :return: (tuple) logger, logport, clients, guis, params
    """

    logger = kwargs['logger']
    loghost = kwargs['loghost']
    clients = kwargs['clients']
    guis = kwargs['guis']
    logport = kwargs['logport']
    params = kwargs['params']

    return logger, loghost, logport, clients, guis, params


def show_console():
    """ Shows the active console.

    Useful for processes where console is typically hidden but user input is suddenly required
    """

    ctypes.windll.user32.ShowWindow(ctypes.windll.kernel32.GetConsoleWindow(), 9)

def hide_console():
    """ Hides the active console.

    Useful for processes where console is not needed (isntead, there is a GUI to use)
    """

    ctypes.windll.user32.ShowWindow(ctypes.windll.kernel32.GetConsoleWindow(), 6)

def create_server(service, logger=None, host='localhost'):
    """ Attempts to create a server with randomly chosen port numbers

    :param service: service from which to launch a server
    :param logger: instance of LogClient for logging
    :param host: (optinal) IP address of host

    :return: (tuple) instance of server created, port number (int) created on
    """

    timeout = 0
    while timeout < 1000:
        try:
            port = np.random.randint(1, 9999)
            server = GenericServer(
                host=host,
                port=port,
                service=service
            )
            timeout = 9999
        except ConnectionRefusedError:
            msg_str = f'Failed to create update server with port {port}'
            if logger is None:
                print(msg_str)
            else:
                logger.warn(f'Failed to create update server with port {port}')
            timeout += 1
    return server, port



import ctypes
from pylabnet.hardware.p_gen.ni_hsdio import NI_HSDIO_DLL_PATH
from pylabnet.hardware.p_gen.ni_hsdio.c_headers import NITypes, NIConst, build_c_func_prototypes
from pylabnet.hardware.interface.p_gen import PGenError
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase
from pylabnet.utils.pulseblock.pb_sample import pb_sample
from pylabnet.utils.logging.logger import LogHandler
import numpy as np
import copy


class Driver:
    """Driver class for NI PXI 654x HSDIO card
    """

    def __init__(self, dev_name_str, logger=None):

        #
        # Define internal variables
        #

        self.log = LogHandler(logger=logger)
        self.map_dict = dict()
        self.writn_wfm_set = set()

        #
        # "Load" niHSDIO DLL
        #

        try:
            self.dll = ctypes.WinDLL(NI_HSDIO_DLL_PATH)
        except OSError:
            msg_str = 'DLL loading failed. \n' \
                      'Ensure that niHSDIO DLL path is correct: \n' \
                      'it should be specified in pylabnet.hardware.p_gen.ni_hsdio.__init__.py \n' \
                      'Current value is: \n' \
                      '"{}"'.format(NI_HSDIO_DLL_PATH)
            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)

        # Build C-prototypes (in particular, specify the return
        # types such that Python reads results correctly)
        build_c_func_prototypes(self.dll)

        #
        # Connect to device
        #

        # Create blank session handle
        self._handle = NITypes.ViSession()

        # Convert dev_name to the DLL-readable format
        self._dev_name = NITypes.ViRsrc(
            dev_name_str.encode('ascii')
        )

        self._er_chk(
            self.dll.niHSDIO_InitGenerationSession(
                self._dev_name,             # ViRsrc resourceName
                NIConst.VI_TRUE,            # ViBoolean IDQuery
                NIConst.VI_TRUE,            # ViBoolean resetDevice
                NIConst.VI_NULL,            # ViConstString optionString - not used, set to VI_NULL
                ctypes.byref(self._handle)  # ViSession * session_handle
            )
        )

        # Log info message
        serial_str = self._get_attr_str(
            NIConst.NIHSDIO_ATTR_SERIAL_NUMBER
        )
        self.log.info(
            'Connected to NI HSDIO card {0}. Serial number: {1}. Session handle: {2}'
            ''.format(dev_name_str, serial_str, self._handle)
        )

    def reset(self):

        self.writn_wfm_set = set()
        self.map_dict = dict()

        return self._er_chk(
            self.dll.niHSDIO_reset(self._handle)
        )

    def start(self):
        return self._er_chk(
            self.dll.niHSDIO_Initiate(
                self._handle
            )
        )

    def stop(self):
        return self._er_chk(
            self.dll.niHSDIO_Abort(
                self._handle
            )
        )

    def disconnect(self):
        self.reset()

        op_status = self._er_chk(
            self.dll.niHSDIO_close(self._handle)
        )
        return op_status

    # ================================================================
    # Hardware settings
    # ================================================================

    def get_samp_rate(self):
        return self._get_attr_real64(
            NIConst.NIHSDIO_ATTR_SAMPLE_CLOCK_RATE
        )

    def set_samp_rate(self, samp_rate):

        # Sanity check
        if not samp_rate <= self.constraints['samp_rate']['max']:
            self.log.warn(
                'set_samp_rate({0} MHz): the requested value exceeds hardware constraint max={1} MHz.\n'
                'The max possible value will be set instead.'
                ''.format(samp_rate / 1e6, self.constraints['samp_rate']['max'] / 1e6)
            )
            samp_rate = self.constraints['samp_rate']['max']

        elif not self.constraints['samp_rate']['min'] <= samp_rate:
            self.log.warn(
                'set_samp_rate({0} Hz): the requested value is below the hardware constraint min={1} Hz.\n'
                'The min possible value will be set instead.'
                ''.format(samp_rate, self.constraints['samp_rate']['min'])
            )
            samp_rate = self.constraints['samp_rate']['min']

        # Call DLL function
        # Currently, the onboard clock is used as the sample clock source
        self._er_chk(
            self.dll.niHSDIO_ConfigureSampleClock(
                self._handle,                            # ViSession vi
                NIConst.NIHSDIO_VAL_ON_BOARD_CLOCK_STR,  # ViConstString clockSource
                NITypes.ViReal64(samp_rate)              # ViReal64 clockRate
            )
        )

        # Return the actual final sample rate
        return self.get_samp_rate()

    def get_active_chs(self):
        return self._get_attr_str(
            NIConst.NIHSDIO_ATTR_DYNAMIC_CHANNELS
        )

    def set_active_chs(self, chs_str=None):

        if chs_str is None:
            # un-assign all channels
            chs_str = 'none'

        chs_str = NITypes.ViString(chs_str.encode('ascii'))

        self._er_chk(
            self.dll.niHSDIO_AssignDynamicChannels(
                self._handle,  # ViSession vi,
                chs_str        # ViConstString channelList
            )
        )

        return self.get_active_chs()

    def get_mode(self):
        """

        :return: (str) "W" - Waveform, "S" - Scripted
        """

        mode_id = self._get_attr_int32(
            attr_id=NIConst.NIHSDIO_ATTR_GENERATION_MODE
        )

        if mode_id == NIConst.NIHSDIO_VAL_WAVEFORM.value:
            return 'W'
        elif mode_id == NIConst.NIHSDIO_VAL_SCRIPTED.value:
            return 'S'
        else:
            msg_str = 'get_mode(): self._get_attr_int32(NIHSDIO_ATTR_GENERATION_MODE) ' \
                      'returned unknown mode_id = {0}'.format(mode_id)
            self.log.error(msg_str)
            raise PGenError(msg_str)

    def set_mode(self, mode_string):
        """

        :param mode_string: (str) "W" - Waveform, "S" - Scripted
        :return: actual run mode string ("W"/"S")
        """

        if mode_string == 'W':
            run_mode = NIConst.NIHSDIO_VAL_WAVEFORM
        elif mode_string == 'S':
            run_mode = NIConst.NIHSDIO_VAL_SCRIPTED
        else:
            msg_str = 'set_mode({0}): invalid argument. Valid values: "W" - Waveform, "S" - scripted. \n' \
                      'Run mode was not changed. Actual run mode string was returned.'.format(mode_string)

            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)

        # Call DLL function
        self._er_chk(
            self.dll.niHSDIO_ConfigureGenerationMode(
                self._handle,  # ViSession vi
                run_mode       # ViInt32 generationMode
            )
        )

        # Return actual run mode
        return self.get_mode()

    @property
    def constraints(self):

        # Total memory size
        # [in samples; one sample contains 32 bits and covers all channels]
        max_wfm_len = self._get_attr_int32(
            attr_id=NIConst.NIHSDIO_ATTR_TOTAL_GENERATION_MEMORY_SIZE
        )

        constr_dict = dict(
            samp_rate=dict(
                min=48,
                max=100e6
            ),
            wfm_len=dict(
                min=2,
                step=2,
                max=max_wfm_len
            )
        )

        return constr_dict

    def get_status(self):

        try:
            # Record current samp_rate to restore it later
            current_rate = self.get_samp_rate()
            rate_lims = self.constraints['samp_rate']

            test_rate = (rate_lims['min'] + rate_lims['max']) / 2

            # Try changing samp_rate
            op_status = self.dll.niHSDIO_ConfigureSampleClock(
                self._handle,                            # ViSession vi
                NIConst.NIHSDIO_VAL_ON_BOARD_CLOCK_STR,  # ViConstString clockSource
                NITypes.ViReal64(test_rate)              # ViReal64 clockRate
            )

            # If device is idle, operation should be successful:
            #   op_status = 0.
            # Restore original samp rate and return 0 - "idle"

            if op_status == 0:
                self.set_samp_rate(samp_rate=current_rate)
                return 0

            # If device is running, attempt to change samp_rate should return
            # the following error code:
            #   -1074118617
            #   "Specified property cannot be set while the session is running.
            #   Set the property prior to initiating the session,
            #   or abort the session prior to setting the property."

            elif op_status == -1074118617:
                # Device is running
                return 1

            # This method cannot interpret any other error/warning code and has
            # to raise an exception

            else:
                raise PGenError(
                    'get_status(): the attempt to test-change samp_rate returned unknown error code {}'
                    ''.format(op_status)
                )

        # If connection to the device is lost
        # or any operation fails, raise an exception.

        except Exception as exc_obj:
            self.log.exception(
                'get_status(): an exception was produced. \n'
                'This might mean that connection to the device is lost '
                'or there is some bug in the get_status() method. \n'
            )
            raise exc_obj

    # ================================================================
    # Waveform Generation
    # ================================================================

    def write_wfm(self, pb_obj, len_adj=True):

        #
        # Sanity checks
        #

        # Only data_width=32 write is currently implemented
        # (DLL function niHSDIO_WriteNamedWaveformU32)
        hrdw_data_width = 8 * self._get_attr_int32(NIConst.NIHSDIO_ATTR_DATA_WIDTH)
        if hrdw_data_width != 32:
            msg_txt = 'write_wfm(): the card you use has data_width = {0} bits. \n' \
                      'The method was written assuming 32-bit width and have to be modified for your card. \n' \
                      'Rewrite bit_ar construction part and use niHSDIO_WriteNamedWaveformU{1}() DLL function' \
                      ''.format(hrdw_data_width, hrdw_data_width)

            self.log.error(msg_txt)
            raise PGenError(msg_txt)

        #
        # Sample PulseBlock
        #

        # Map user-friendly names onto physical channel numbers
        pb_obj = copy.deepcopy(pb_obj)
        pb_obj.ch_map(map_dict=self.map_dict)

        # Sample pulse block
        samp_rate = self.get_samp_rate()
        samp_dict, n_pts, add_pts = pb_sample(
            pb_obj=pb_obj,
            samp_rate=samp_rate,
            len_min=self.constraints['wfm_len']['min'],
            len_max=self.constraints['wfm_len']['max'],
            len_step=self.constraints['wfm_len']['step'],
            len_adj=len_adj
        )
        wfm_name = pb_obj.name
        del pb_obj

        self.log.info(
            'write_wfm(): sampled PulseBlock "{}". \n'
            'Sample array has {} points. {} samples were added to match hardware wfm len step'
            ''.format(wfm_name, n_pts, add_pts)
        )

        #
        # Pack samp_dict into bit_ar
        #

        # Create a blank bit_ar - all elements zero (all channels off)
        bit_ar = np.zeros(shape=n_pts, dtype=np.uint32)

        # Iterate through each channel and set corresponding bit to '1'
        # if value is True
        for ch_num in samp_dict.keys():

            # This number in uint32 representation has all zeros and
            # exactly one '1' at the ch_num-th bit from the LSB
            ch_bit_one = 2**ch_num

            for idx, val in enumerate(samp_dict[ch_num]):
                if val:
                    bit_ar[idx] += ch_bit_one

            # TODO: consider making very fast with numpy:
            # ch_bit_ar = samp_dict[ch_num].astype(int) * 2**ch_num
            # bit_ar = np.add(bit_ar, ch_bit_ar)

        #
        # Load bit_ar to memory
        #

        # Delete waveform with the same name,
        # if it is already present in the memory
        if wfm_name in self.writn_wfm_set:
            self.del_wfm(wfm_name=wfm_name)

        # Create C-pointer to bit_ar using numpy.ndarray.ctypes attribute
        bit_ar_ptr = bit_ar.ctypes.data_as(
            ctypes.POINTER(NITypes.ViUInt32)
        )

        # Call DLL function
        self._er_chk(
            self.dll.niHSDIO_WriteNamedWaveformU32(
                self._handle,                                     # ViSession vi
                NITypes.ViConstString(wfm_name.encode('ascii')),  # ViConstString waveformName
                NITypes.ViInt32(n_pts),                           # ViInt32 samplesToWrite
                bit_ar_ptr                                        # ViUInt32 data[]
            )
        )

        self.writn_wfm_set.add(wfm_name)

        return 0

    def del_wfm(self, wfm_name):
        self._er_chk(
            self.dll.niHSDIO_DeleteNamedWaveform(
                self._handle,                                    # ViSession vi
                NITypes.ViConstString(wfm_name.encode('ascii'))  # ViConstString waveformName
            )
        )

        self.writn_wfm_set.remove(wfm_name)

        return 0

    def clr_mem(self):
        wfm_set = copy.deepcopy(self.writn_wfm_set)

        for wfm_name in wfm_set:
            self.del_wfm(wfm_name=wfm_name)

        return 0

    def get_rep(self):
        """Returns number of repetitions in Waveform generation mode.

        On the hardware level, it is just a pair of attributes
        NIHSDIO_ATTR_REPEAT_MODE and NIHSDIO_ATTR_REPEAT_COUNT
        which are not bound to any specific waveform.

        :return: (int) repeat mode + number of repetitions:
                 0 - repeat infinitely
                 positive integer - finite, number of repetitions
                 PGenError exception - error
        """

        rep_mode = self._get_attr_int32(
            NIConst.NIHSDIO_ATTR_REPEAT_MODE
        )

        if rep_mode == NIConst.NIHSDIO_VAL_CONTINUOUS.value:
            rep_num = 0
        elif rep_mode == NIConst.NIHSDIO_VAL_FINITE.value:
            rep_num = self._get_attr_int32(
                NIConst.NIHSDIO_ATTR_REPEAT_COUNT
            )
        else:
            msg_str = 'get_rep(): DLL call returned unknown repetition mode code {}'.format(rep_mode)

            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)

        return rep_num

    def set_rep(self, rep_num):
        """Set repeat mode and number of repetitions

        :param rep_num: (int) repeat mode + number of repetitions:
                    0 - repeat infinitely
                    positive integer - finite, number of repetitions

        :return: (int) actual repeat mode + number of repetitions:
                    0 - repeat infinitely
                    positive integer - finite, number of repetitions
                    PGenError exception - error
        """

        if rep_num == 0:
            rep_mode = NIConst.NIHSDIO_VAL_CONTINUOUS
            rep_num = NIConst.VI_NULL

        elif rep_num > 0:
            rep_mode = NIConst.NIHSDIO_VAL_FINITE
            rep_num = NITypes.ViInt32(rep_num)

        else:
            msg_str = 'set_rep() invalid argument {} \n' \
                      'Valid values: 0 - infinite, positive integer - finite' \
                      ''.format(rep_num)

            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)

        self._er_chk(
            self.dll.niHSDIO_ConfigureGenerationRepeat(
                self._handle,  # ViSession vi
                rep_mode,      # ViInt32 repeatMode
                rep_num        # ViInt32 repeatCount
            )
        )

        return self.get_rep()

    def get_wfm_to_gen(self):
        return self._get_attr_str(
            NIConst.NIHSDIO_ATTR_WAVEFORM_TO_GENERATE
        )

    def set_wfm_to_gen(self, wfm_name):
        self._er_chk(
            self.dll.niHSDIO_ConfigureWaveformToGenerate(
                self._handle,                                    # ViSession vi
                NITypes.ViConstString(wfm_name.encode('ascii'))  # ViConstString waveformName
            )
        )
        return self.get_wfm_to_gen()

    # ================================================================
    # Script Generation
    # ================================================================

    def write_script(self, script_str):

        # Sanity check: script_str is a string
        if not isinstance(script_str, str):
            msg_str = 'write_script(): passed argument is not a string'

            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)

        plain_script_str = script_str.replace('\n', ' ')

        # Convert into C-string
        c_script_str = NITypes.ViConstString(
            plain_script_str.encode('ascii')
        )

        op_status = self._er_chk(
            self.dll.niHSDIO_WriteScript(
                self._handle,  # ViSession vi
                c_script_str   # ViConstString script
            )
        )

        return op_status

    def get_scr_to_gen(self):
        return self._get_attr_str(
            NIConst.NIHSDIO_ATTR_SCRIPT_TO_GENERATE
        )

    def set_scr_to_gen(self, script_name):

        # Convert script_name into C-string
        script_name = NITypes.ViConstString(
            script_name.encode('ascii')
        )

        self._er_chk(
            self.dll.niHSDIO_ConfigureScriptToGenerate(
                self._handle,  # ViSession vi
                script_name    # ViConstString scriptName
            )
        )

        return self.get_scr_to_gen()

    # ================================================================
    # Wrappers for C DLL helper functions
    # ================================================================

    def _er_chk(self, error_code):
        # C:\Program Files\National Instruments\Shared\Errors\English\IVI-errors.txt

        if error_code == 0:
            # Success
            return 0
        else:
            # Warning or Error

            # Create buffer for DLL function to output the error message string
            msg_buf = ctypes.create_string_buffer(256)

            # Call DLL function to generate readable error message
            self.dll.niHSDIO_error_message(
                self._handle,                  # ViSession vi
                NITypes.ViStatus(error_code),  # ViStatus errorCode
                msg_buf                        # ViChar errorMessage[256]
            )
            msg_str = msg_buf.value.decode('ascii')

            if error_code > 0:
                # Warning
                self.log.warn(msg_str=msg_str)
                return error_code
            else:
                # Error
                self.log.error(msg_str=msg_str)
                raise PGenError(msg_str)

    def _get_attr_int32(self, attr_id, ch=None):
        """

        :param attr_id:
        :param ch: (int)
        :return: (int) obtained value in the case of success
                 PGenError exception is produced in the case of error
        """

        # Create buffer where niHSDIO_GetAttribute will store the result
        buf = NITypes.ViInt32()

        # Convert channel number into C-string
        # (used to request channel-agnostic/-specific attributes)
        if ch is None:
            ch_str = NIConst.VI_NULL
        else:
            ch_str = str(ch)
            # Convert into C-string
            ch_str = ctypes.c_char_p(ch_str.encode('ascii'))

        # Call DLL function
        try:
            self._er_chk(
                self.dll.niHSDIO_GetAttributeViInt32(
                    self._handle,      # ViSession vi
                    ch_str,            # ViConstString channelName
                    attr_id,           # ViAttr attribute
                    ctypes.byref(buf)  # ViInt32 *value
                )
            )
            return buf.value

        except OSError:
            # DLL normally handles all "moderate" errors and returns error code,
            # which is being analyzed by self._er_chk.
            # "try" handles OSError when the DLL function fails completely
            # and isn't able to handle the error by itself

            msg_str = '_get_attr_int32(): OSError, DLL function call failed'
            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)

    def _get_attr_str(self, attr_id, ch=None):
        """

        :param attr_id:
        :param ch: (int)
        :return: (str) obtained value in the case of success
                       Exception is produced in the case of error
        """

        # Create buffer where niHSDIO_GetAttribute will store the result
        buf_size = 1024  # Buffer size of 1024 was chosen for no specific reason. Increase if needed.
        buf = ctypes.create_string_buffer(buf_size)

        # Convert channel number into C-string
        # (used to request channel-agnostic/-specific attributes)
        if ch is None:
            ch_str = NIConst.VI_NULL
        else:
            ch_str = str(ch)
            # Convert into C-string
            ch_str = ctypes.c_char_p(ch_str.encode('ascii'))

        # Call DLL function
        try:
            self._er_chk(
                self.dll.niHSDIO_GetAttributeViString(
                    self._handle,               # ViSession vi
                    ch_str,                     # ViConstString channelName
                    attr_id,                    # ViAttr attribute
                    NITypes.ViInt32(buf_size),  # ViInt32 bufSize
                    buf                         # ViChar value[]
                )
            )
            return buf.value.decode('ascii')

        except OSError:
            # DLL normally handles all "moderate" errors and returns error code,
            # which is being analyzed by self._er_chk.
            # "try" handles OSError when the DLL function fails completely
            # and isn't able to handle the error by itself

            msg_str = '_get_attr_str(): OSError, DLL function call failed'
            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)

    def _get_attr_real64(self, attr_id, ch=None):
        """

        :param attr_id:
        :param ch: (int)
        :return: (float) obtained value in the case of success
                 Exception is produced in the case of error
        """

        # Create buffer where niHSDIO_GetAttribute will store the result
        buf = NITypes.ViReal64()

        # Convert channel number into C-string
        # (used to request channel-agnostic/-specific attributes)
        if ch is None:
            ch_str = NIConst.VI_NULL
        else:
            ch_str = str(ch)
            # Convert into C-string
            ch_str = ctypes.c_char_p(ch_str.encode('ascii'))

        # Call DLL function
        try:
            self._er_chk(
                self.dll.niHSDIO_GetAttributeViReal64(
                    self._handle,      # ViSession vi
                    ch_str,            # ViConstString channelName
                    attr_id,           # ViAttr attribute
                    ctypes.byref(buf)  # ViReal64 *value
                )
            )
            return buf.value

        except OSError:
            # DLL normally handles all "moderate" errors and returns error code,
            # which is being analyzed by self._er_chk.
            # "try" handles OSError when the DLL function fails completely
            # and isn't able to handle the error by itself

            msg_str = '_get_attr_real64(): OSError, DLL function call failed'
            self.log.error(msg_str=msg_str)
            raise PGenError(msg_str)


class Service(ServiceBase):
    # TODO: implement
    pass


class Client(ClientBase):
    # TODO: implement
    pass



import abc


class CtrMonitorInterface(abc.ABC):

    @abc.abstractmethod
    def start_counting(self, bin_width=1000000000, n_bins=10000):
        """

        :param bin_width: Width of counting bins
        :param n_values: Number of values of counter array
        """

        pass

    @abc.abstractmethod
    def clear_counter(self):
        pass

    @abc.abstractmethod
    def get_counts(self):
        pass

    @abc.abstractmethod
    def get_x_axis(self):
        pass

        """

        :param ch_list: List of channels to count
        """

        pass

    @abc.abstractmethod
    def set_channels(self, ch_list=[1]):
        """

        :param ch_list: List of channels to use [1,8] for
            rising edges and [-8,-1] for falling edges
        """

        pass
    

class CtrError(Exception):
    pass



import visa
from pylabnet.utils.logging.logger import LogHandler
from pylabnet.hardware.interface.mw_src import MWSrcInterface, MWSrcError
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase
import pickle


class Driver(MWSrcInterface):
    """Adapted from Qudi <https://github.com/Ulm-IQO/qudi/>
    """

    def __init__(self, addr_str, logger=None):

        self.log = LogHandler(logger=logger)

        # Connect to the device
        self.rm = visa.ResourceManager()
        try:
            self._dev = self.rm.open_resource(addr_str)
        except Exception as exc_obj:
            self.log.exception(
                msg_str='Could not connect to the address >>{}<<. \n'.format(addr_str)
            )
            raise exc_obj

        # Log confirmation info message
        id_str = self._dev.query('*IDN?').replace(',', ' ')
        id_str = id_str.strip('\n')
        self.log.info(
            msg_str='{} initialised and connected.'.format(id_str)
        )

        # Reset device
        self.reset()

        # Error check
        self._er_chk()

    def reset(self):
        # Reset
        self._cmd_wait('*RST')
        # Clear status register
        self._cmd_wait('*CLS')

        return 0

    def activate_interface(self):

        # Store hardware settings which are not controlled by logic,
        # to restore them after reset()
        # [logic does not know anything about this params, so it should not
        # introduce any changes to them by calling activate_interface()].
        tmp_trig_dict = self.get_trig()

        # Reset device
        self.reset()

        # Restore hardware settings which are not controlled by logic
        # but were changed by self._dev.reset()

        self.set_trig(
            src_str=tmp_trig_dict['src_str'],
            slope_str=tmp_trig_dict['slope_str']
        )

        return 0

    # Output control

    def on(self):

        if self.get_mode() == 'sweep':
            self.reset_swp_pos()

        return self._cmd_wait(
            cmd_str=':OUTP:STAT ON'
        )

    def off(self):
        return self._cmd_wait(
            cmd_str=':OUTP:STAT OFF'
        )

    def get_status(self):
        status = int(
            self._dev.query('OUTP:STAT?')
        )
        return status

    # Power

    def get_pwr(self):
        return float(self._dev.query(':POW?'))

    def set_pwr(self, pwr):
        self._cmd_wait(':POW {0:f}'.format(pwr))

        return self.get_pwr()

    # Frequency

    def get_freq(self):
        mode = self.get_mode()

        if mode == 'cw':
            ret_val = float(self._dev.query(':FREQ?'))

        elif mode == 'sweep':
            start = float(self._dev.query(':FREQ:STAR?'))
            stop = float(self._dev.query(':FREQ:STOP?'))
            step = float(self._dev.query(':SWE:STEP?'))

            n_pts = int((stop - start) / step) + 2

            ret_val = dict(
                start=start,
                stop=stop,
                n_pts=n_pts
            )

        else:
            raise MWSrcError(
                'get_freq(): got unknown mode {}'.format(mode)
            )

        return ret_val

    def set_freq(self, freq):

        if self.get_status() == 1:
            self.off()

        # Activate CW mode
        self._cmd_wait(':FREQ:MODE CW')

        # Set CW frequency
        self._cmd_wait(':FREQ {0:f}'.format(freq))

        return self.get_freq()

    def set_freq_swp(self, start, stop, n_pts):

        if self.get_status() == 1:
            self.off()

        # Set mode to Sweep
        self._cmd_wait(':FREQ:MODE SWEEP')

        # Set frequency sweep
        step = (stop - start) / (n_pts - 1)

        self._cmd_wait(':SWE:MODE STEP')
        self._cmd_wait(':SWE:SPAC LIN')
        self._cmd_wait(':FREQ:START {0:f}'.format(start))
        self._cmd_wait(':FREQ:STOP {0:f}'.format(stop))
        self._cmd_wait(':SWE:STEP:LIN {0:f}'.format(step))

        return self.get_freq()

    def reset_swp_pos(self):
        """Reset of MW sweep mode position to start (start frequency)

        @return int: error code (0:OK, -1:error)
        """

        self._cmd_wait(':ABOR:SWE')
        return 0

    def get_mode(self):

        mode_str = self._dev.query(':FREQ:MODE?').strip('\n').lower()

        if 'cw' in mode_str:
            return 'cw'
        elif 'swe' in mode_str:
            return 'sweep'
        else:
            msg_str = 'get_mode(): unknown mode string {} was returned'
            self.log.error(msg_str=msg_str)
            raise MWSrcError(msg_str)

    # Technical methods

    def _cmd_wait(self, cmd_str):
        """Writes the command in command_str via resource manager
        and waits until the device has finished processing it.

        @param cmd_str: The command to be written
        """

        self._dev.write(cmd_str)

        # Block command queue until cmd_str execution is complete
        self._dev.write('*WAI')
        # Block Python process until cmd_str execution is complete
        self._dev.query('*OPC?')

        # Error check
        self._er_chk()

        return 0

    def _er_chk(self):
        # Block command queue until all previous commands are complete
        self._dev.write('*WAI')
        # Block Python process until all previous commands are complete
        self._dev.query('*OPC?')

        # Read all messages
        out_str = self._dev.query(':SYSTem:ERRor:ALL?')
        out_str += ','
        out_str += self._dev.query('SYST:SERR?')

        out_str = out_str.replace('\n', '')
        out_str = out_str.replace('\r', '')
        out_str = out_str.replace('"', '')

        out_list = out_str.split(',')

        # Collect all warns and errors
        er_list = []
        warn_list = []

        msg_n = int(len(out_list) / 2)

        for idx in range(msg_n):
            msg_code = int(out_list[2 * idx])

            if msg_code == 0:
                # No error
                continue

            elif msg_code > 0:
                # Warning
                warn_list.append(out_list[2 * idx + 1])

            else:
                # Error
                er_list.append(out_list[2 * idx + 1])

        # Construct Warn message string
        if len(warn_list) > 0:
            warn_str = ''
            for warn in warn_list:
                warn_str += (warn + ' \n')
            warn_str = warn_str.rstrip('\n')
            self.log.warn(msg_str=warn_str)

        # Construct Error message string
        if len(er_list) > 0:
            er_str = ''
            for er in er_list:
                er_str += (er + ' \n')
            er_str = er_str.rstrip('\n')
            self.log.error(msg_str=er_str)
            raise MWSrcError(er_str)

        return 0

    def get_trig(self):

        #
        # Get trigger source
        #

        src_str = self._dev.query('TRIG:FSW:SOUR?')

        if 'EXT' in src_str:
            src_str = 'ext'
        elif 'AUTO' in src_str:
            src_str = 'int'
        else:
            msg_str = 'get_trig(): unknown trigger source was returned "{}" \n' \
                      ''.format(src_str)
            self.log.error(msg_str=msg_str)
            raise MWSrcError(msg_str)

        #
        # Get edge slope
        #

        slope_str = self._dev.query(':TRIG1:SLOP?')

        if 'POS' in slope_str:
            slope_str = 'r'
        elif 'NEG' in slope_str:
            slope_str = 'f'
        else:
            msg_str = 'get_trig(): unknown slope was returned "{}" \n' \
                      ''.format(slope_str)
            self.log.error(msg_str=msg_str)
            raise MWSrcError(msg_str)

        return dict(src_str=src_str, slope_str=slope_str)

    def set_trig(self, src_str='ext', slope_str='r'):

        if self.get_status() == 1:
            self.off()

        #
        # Set trigger source
        #

        if src_str == 'ext':
            src_str = 'EXT'
        elif src_str == 'int':
            src_str = 'AUTO'
        else:
            msg_str = 'set_trig(): unknown trigger source  "{}" \n' \
                      'Valid values are "ext" - external, "int" - internal' \
                      ''.format(src_str)
            self.log.error(msg_str=msg_str)
            raise MWSrcError(msg_str)

        self._cmd_wait('TRIG:FSW:SOUR {}'.format(src_str))

        #
        # Set trigger edge
        #

        if slope_str == 'r':
            edge = 'POS'
        elif slope_str == 'f':
            edge = 'NEG'
        else:
            msg_str = 'set_trig(): invalid argument slope_str={} \n' \
                      'Valid values are: "r" - raising, "f" - falling' \
                      ''.format(slope_str)
            self.log.error(msg_str=msg_str)
            raise ValueError(msg_str)

        self._cmd_wait(':TRIG1:SLOP {0}'.format(edge))

        return self.get_trig()

    def force_trig(self):
        """ Trigger the next element in the list or sweep mode programmatically.

        @return int: error code (0:OK, -1:error)

        Ensure that the Frequency was set AFTER the function returns, or give
        the function at least a save waiting time.
        """

        self._cmd_wait('*TRG')

        return 0


class Service(ServiceBase):

    def exposed_activate_interface(self):
        return self._module.activate_interface()

    # Output control

    def exposed_on(self):
        return self._module.on()

    def exposed_off(self):
        return self._module.off()

    def exposed_get_status(self):
        return self._module.get_status()

    # Power

    def exposed_get_pwr(self):
        return self._module.get_pwr()

    def exposed_set_pwr(self, pwr):
        return self._module.set_pwr(pwr=pwr)

    # Frequency

    def exposed_get_freq(self):
        res = self._module.get_freq()
        return pickle.dumps(res)

    def exposed_set_freq(self, freq):
        return self._module.set_freq(freq=freq)

    def exposed_set_freq_swp(self, start, stop, n_pts):
        res = self._module.set_freq_swp(
            start=start,
            stop=stop,
            n_pts=n_pts
        )
        return pickle.dumps(res)

    def exposed_reset_swp_pos(self):
        return self._module.reset_swp_pos()

    def exposed_get_mode(self):
        return self._module.get_mode()


class Client(ClientBase, MWSrcInterface):

    def activate_interface(self):
        return self._service.exposed_activate_interface()

    # Output control

    def on(self):
        return self._service.exposed_on()

    def off(self):
        return self._service.exposed_off()

    def get_status(self):
        return self._service.exposed_get_status()

    # Power

    def get_pwr(self):
        return self._service.exposed_get_pwr()

    def set_pwr(self, pwr):
        return self._service.exposed_set_pwr(
            pwr=float(pwr)
        )

    # Frequency

    def get_freq(self):
        ret_pckl = self._service.exposed_get_freq()
        return pickle.loads(ret_pckl)

    def set_freq(self, freq):
        return self._service.exposed_set_freq(
            freq=float(freq)
        )

    def set_freq_swp(self, start, stop, n_pts):
        res_pckl = self._service.exposed_set_freq_swp(
            start=start,
            stop=stop,
            n_pts=n_pts
        )

        return pickle.loads(res_pckl)

    def reset_swp_pos(self):
        return self._service.exposed_reset_swp_pos()

    def get_mode(self):
        return self._service.exposed_get_mode()




"""SequenceGenerator wrapper for Tektronix 7k AWG"""

import copy


class Wrap:

    def __init__(self, tek_awg_7k_inst):
        self._dev = tek_awg_7k_inst

        # Channel name map: user-friendly to device-native
        self.map_dict = None

        self.reset()

    def reset(self):
        """ Interface-level reset

        Resets everything, except for hardware settings to which
        the above-lying logic has no access.

        :return: 0 - Success code
        """

        # Save all the settings, which were not given by above-lying logic
        tmp_interleave = self._dev.get_interleave()
        tmp_samp_rate = self._dev.get_samp_rate()
        tmp_anlg_level = self._dev.get_analog_level()
        tmp_digital_level = self._dev.get_digital_level()

        # Full hardware reset
        self._dev.reset()

        # Restore these settings
        self._dev.set_mode(mode_string='S')
        self._dev.set_interleave(state=tmp_interleave)
        self._dev.set_samp_rate(samp_rate=tmp_samp_rate)
        self._dev.set_analog_level(level_dict=tmp_anlg_level)
        self._dev.set_digital_level(level_dict=tmp_digital_level)

        return 0

    def start(self):
        return self._dev.start()

    def stop(self):
        return self._dev.stop()

    def write_wfm(self, pb_obj, collapse=False, strict_hrdw_seq=False, len_adj=True):

        # Map user-friendly names onto physical channel numbers
        pb_obj = copy.deepcopy(pb_obj)
        pb_obj.ch_map(map_dict=self.map_dict)

        # Fill unused channels with default values
        pb_obj = self._dev.fill_unused_chs(pb_obj=pb_obj)

        if collapse:
            self._dev.write_wfm_zip(
                pb_obj=pb_obj,
                len_adj=len_adj
            )
        else:
            self._dev.write_wfm(
                pb_obj=pb_obj,
                len_adj=len_adj,
                strict_hrdw_seq=strict_hrdw_seq
            )

    def del_wfm(self, wfm_name):
        pass

    def write_seq(self, seq_table):
        pass





import pickle
import numpy as np
import matplotlib.pyplot as plt

from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class Service(ServiceBase):

    def exposed_reset(self):
        return self._module.reset()

    def exposed_display_off(self):
        return self._module.display_off()

    def exposed_display_on(self):
        return self._module.display_on()

    def exposed_set_attenuation(self, db):
        return self._module.set_attenuation(db)

    def exposed_set_center_frequency(self, center_frequency):
        return self._module.set_center_frequency(center_frequency)

    def exposed_set_frequency_span(self, frequency_span):
        return self._module.set_frequency_span(frequency_span)

    def exposed_read_trace(self):
        trace = self._module.read_trace()
        return pickle.dumps(trace)

    def exposed_query(self, command):
        query = self._module.device.query(command)
        return pickle.dumps(query)

    def exposed_write(self, command):
        return self._module.device.write(command)

    def exposed_acquire_background_spectrum(self, num_point):
        return self._module.acquire_background_spectrum(num_point)

    def exposed_get_background(self):
        background = self._module.get_background()
        return pickle.dumps(background)


class Client(ClientBase):

    def reset(self):
        return self._service.exposed_reset()

    def display_off(self):
        return self._service.exposed_display_off()

    def display_on(self):
        return self._service.exposed_display_on()

    def set_attenuation(self, db):
        return self._service.exposed_set_attenuation(db)

    def set_center_frequency(self, center_frequency):
        return self._service.exposed_set_center_frequency(center_frequency)

    def set_frequency_span(self, frequency_span):
        return self._service.exposed_set_frequency_span(frequency_span)

    def read_trace(self):
        pickled_trace = self._service.exposed_read_trace()
        return pickle.loads(pickled_trace)

    def write(self, command):
        return self._service.exposed_write(command)

    def query(self, command):
        pickled_query = self._service.exposed_query(command)
        return pickle.loads(pickled_query)

    def acquire_background_spectrum(self, num_point=100, nocheck=True):

        if not nocheck:
            print("Please make sure that spectrum analyzer is measuring noise floor, i.e. no input signals are switched on. Type 'y' to confirm.")
            a = input()
            if a == 'y':
                return self._service.acquire_background_spectrum(num_point)
            else:
                print('Wrong input, returns.')
                return

    def get_background(self):
        pickled_background = self._service.exposed_get_background()

        background = pickle.loads(pickled_background)

        if not type(background) == np.ndarray:
            print('No background acquired. Run acquire_background_spectrum() to acquire background.')

        return pickle.loads(pickled_background)

    def plot_trace(self, trace=None, background_substract=False):
        """Reads and plots trace

        :trace: Trace to plot. If none, acquire new trace.
        :background_substract: If True, substract background from trace
        """

        # If not trace furnished, acquired trace.
        if trace is None:
            trace = self.read_trace()

        # Substract background
        if background_substract:

            # Try to retrieve background
            background = self.get_background()

            # Check if background was successfully retrieved
            if not type(background) == np.ndarray:
                print()
                return

            trace[:, 1] = trace[:, 1] - background[:, 1]
            ylabel = 'Power ratio above noise floor [dB]'
        else:
            ylabel = 'Power [dBm]'

        # Plot trace
        plt.figure()
        plt.plot(trace[:, 0]/1e9, trace[:, 1])
        plt.xlabel('Frequency [GHz]')
        plt.ylabel(ylabel)
        plt.show()

    def plot_background(self):
        """Reads and plots background trace"""

        # Try to retrieve background.
        background = self.get_background()

        # Check if background was successfully retrieved.
        if not type(background) == np.ndarray:
            print('No background acquired. Run acquire_background_spectrum() to acquire background.')
            return

        self.plot_trace(background, background_substract=False)



from .decorator_utils import get_signature


def handle_gui_errors(func):
    """ Error handling decorator checking all GUI configuration errors

    Checks for EOF error and sets self.gui_connected accordingly

    Can only be added to member function of a gui_handler instance
    """

    def wrapper(self, *args, **kwargs):

        # Check 1: KeyError
        updated = False
        timeout = 5  # Number of times to try function execution before timeout.
        try_num = 0

        # Try until timeout is reached
        while not updated and try_num < timeout:
            try_num += 1
            try:
                updated = True
                self.gui_connected = True
                return func(self, *args, **kwargs)
            except KeyError:
                updated = False

                # Note that we should use a warning here, since KeyErrors can happen during normal operation
                self.logger_client.warn(f"KeyError in calling {func.__name__}({get_signature(*args, **kwargs)}), trying again {try_num}/{timeout}.")

            except EOFError:
                updated = False
                self.gui_connected = False
                self.logger_client.error(f"Gui disconnected for function {func.__name__}({get_signature(*args, **kwargs)}), trying again {try_num}/{timeout}.")

            # Handle case where we run out of GUI elements
            except IndexError:
                updated = False
                self.gui_connected = False
                self.logger_client.error(f"No more room at GUI for function {func.__name__}({get_signature(*args, **kwargs)}), trying again {try_num}/{timeout}.")

            # Handle case where we tried to assign some GUI widget that didn't exist
            except AttributeError:
                updated = False
                self.gui_connected = False
                self.logger_client.error(f"Incorrect GUI widget name for function {func.__name__}({get_signature(*args, **kwargs)}), trying again {try_num}/{timeout}.")

            except ConnectionRefusedError:
                updated = False
                self.gui_connected = False
                self.logger_client.error(f"GUI connection failed for function {func.__name__}({get_signature(*args, **kwargs)}), trying again {try_num}/{timeout}.")

    return wrapper



""" Qudi module to make confocal refocus network-visible

Currently broken, as the link from pylabnet Qudi has been severed
"""


# # Imports of qudi's modules
# from network.core.module import Base, ConfigOption, Connector
# from logic.generic_logic import GenericLogic

# # Import of pylabnet's modules
# # (make sure that pylabnet location is in qudi's sys.path)
# from pylabnet.network.core.generic_server import GenericServer
# from pylabnet.network.core.service_base import ServiceBase


# class RefocusService(ServiceBase):

#     _optimizer_logic = None
#     _scanning_logic = None

#     def set_logic(self, optim_logic, scan_logic):
#         self._optimizer_logic = optim_logic
#         self._scanning_logic = scan_logic

#     def exposed_refocus(self):
#         crosshair_pos = self._scanning_logic.get_position()
#         self._optimizer_logic.start_refocus(crosshair_pos, 'confocalgui')
#         return 0


# class RefocusServer(GenericLogic):
#     """
#     This class should be instantiated as Qudi Logic module
#     to allow starting refocus through a client

#     Config:

#     logic:
#         refocus_server:
#             module.Class: 'refocus_server.RefocusServer'
#             host: 'localhost'
#             port: 1234
#             connect:
#                 optimizer_logic: 'optimizerlogic'
#                 scanner_logic: 'scannerlogic'
#     """

#     _modclass = 'RefocusServer'
#     _modtype = 'logic'

#     _host = ConfigOption(name='host', missing='error')
#     _port = ConfigOption(name='port', missing='error')

#     optimizer_logic = Connector(interface='DoesNotMatter')
#     scanner_logic = Connector(interface='DoesNotMatter')

#     def on_activate(self):
#         """ Initialisation performed during activation of the module.
#         """

#         self._service = RefocusService()
#         self._service.set_logic(
#             optim_logic=self.optimizer_logic(),
#             scan_logic=self.scanner_logic()
#         )

#         self._server = GenericServer(
#             host=self._host,
#             port=self._port,
#             service=self._service
#         )

#         self._server.start()

#     def on_deactivate(self):
#         """ Deinitialisation performed during deactivation of the module.
#         """
#         pass




import pulseblock.pulse_block as pb
import numpy as np
import copy


def pb_zip(pb_obj, dur_quant):
    """ Collapse default (wait) periods into repetitions of a single wait block.

    This method takes a plain PulseBlock and returns a sequence table and a list
    of individual waveforms (snippets).
    Corresponding pulse sequence reproduces the original waveform but with long
    wait periods replaced by multiple repetitions of the same short wait element
    for saving hardware memory (collapsing or zipping of a waveform).

    :param pb_obj: PulseObject instance - original pulse block to be collapsed

    :param dur_quant: (float) minimal waveform duration allowed by the hardware
    memory (in the same units as used in pb_obj).
    This value is normally determined as
        min_wfm_size / sampling_rate,
    where min_wfm_size is the minimal size of a sample array in hardware memory.
    The repeated wait waveform will have this size.

    :return: (dict)
    {
        'seq_list': list of tuples ('wfm_name', repetition_number) - chronological
        sequence of (name of the wfm snippet, how many times it should be repeated)

        'snip_list': list of PulseBlock objects - 'snippets' of the original pb_obj,
        which form the sequence according to seq_list
        Naming convention:
            pb_obj.name + '_wait' for the repeated wait wfm
            pb_obj.name + '_' + 'snip number {0, 1, ...}' for non-repeated snippets
        Order: wait is the first element, non-repeated snippets go next in
            chronological order.
    }
    """

    # Non-default intervals ---------------------------------------------------

    # find all time intervals covered by non-default pulses
    p_interval_list = []
    for ch in pb_obj.p_dict.keys():
        for pulse in pb_obj.p_dict[ch]:
            p_interval_list.append(
                (pulse.t0, pulse.t0 + pulse.dur)
            )

    # merge all overlapping pulse intervals
    p_interval_list = merge_intervals(i_list=p_interval_list)

    # Step usage dictionary ---------------------------------------------------

    # Now one knows all the wait periods witch can be collapsed
    # (full duration of pb_obj minus all pulse intervals)
    #
    # But both the wait waveform and non-trivial pulse waveforms have to
    # meet the length constraint: dur_quant + n*dur_step. That is why it is
    # not trivial to split full duration into repetitions of wait element and
    # non-default waveforms.
    #
    # Here one uses the easiest (perhaps not the most efficient) way to meet
    # these restrictions:
    # - divide the entire duration of pb_obj into discrete steps of dur_quant
    # - construct dur_quant-long wait waveform
    # - if a given step does not contain any pulses (or any parts of pulses),
    # it will be replaced by a repetiotion of wait wfm
    # - if one or several adjacent steps contain non-default pulses, they are
    # merged and output as a single pulse block
    # - last integer dur_quant is merged with the reminder (shorter than dur_quant)
    # and is returned as a single pulse block (because the reminder is shorter
    # than minimal waveform length and cannot be sampled alone)

    # Number of whole dur_quant, fitting into pb_obj.dur
    quant_num = int(pb_obj.dur // dur_quant)

    # Use array:
    #   False - this dur_quant is a wait period
    #   True - this dur_quant is a part of pulse interval
    use_ar = np.full(
        shape=quant_num,
        fill_value=False
    )

    # Determine which steps contain non-default pulses
    for p_interval in p_interval_list:
        left_idx = int(p_interval[0] // dur_quant)
        right_idx = int(p_interval[1] // dur_quant)

        # Handle left-most interval
        #   If p_interval[0] == 0.0, integer division by dur_quant
        #   may accidentally evaluate to -1.0
        if left_idx < 0:
            left_idx = 0

        # Handle right-most intervals
        if left_idx == quant_num-1 or left_idx == quant_num:
            # The entire interval fits into the 'default-True' area
            # [(quant_num - 1)-st element and reminder], so no change
            # has to be made in use_ar
            continue
        elif right_idx == quant_num:
            # The left edge of the interval is outside of the 'default-True'
            # area [(quant_num - 1)-st elem and reminder], but the right edge
            # is in the reminder. The pulse might introduce non-trivial change
            # into use_ar, but right_idx is currently out of range.
            #
            # Reduce it to the max available index to avoid out-of-range and
            # broadcasting errors:
            right_idx = quant_num - 1

        use_ar[left_idx : right_idx+1] = np.full(
            shape=right_idx - left_idx + 1,
            fill_value=True
        )
    # Last integer step of dur_quant must be considered as non-wait
    # (to be merged and sampled with the reminder, which is shorter
    # than the hardware minimum)
    use_ar[-1] = True

    # Find periods of pulses / repeating wait (perform 'run-length encoding'):

    # The return is a period dictionary:
    # {
    #   'val_ar': True or False - this run of steps is non-default or wait
    #   'len_ar': number of dur_quant steps within each run
    #   'start_ar': starting index of each run
    # }
    period_dict = run_len_encode(in_ar=use_ar)

    # Final sequence table and PulseBlock dictionary --------------------------

    final_snip_list = []
    final_seq_list = []

    # Construct 'wait' PulseBlock
    wait_pb_name = pb_obj.name + '_wait'
    wait_pb = pb.PulseBlock(name=wait_pb_name)
    wait_pb.dur = dur_quant
    wait_pb.dflt_dict = copy.deepcopy(pb_obj.dflt_dict)
    final_snip_list.append(wait_pb)

    # Iterate through all pulse runs, take 'snippet pulse blocks',
    # and construct sequence table

    # The last run (must be non-wait) will be extended to include the reminder
    last_period_idx = len(period_dict['val_ar']) - 1
    # This counter is used to keep track of PulseBlock names
    pb_name_idx = 0

    for period_idx, period_val in enumerate(period_dict['val_ar']):

        # Pulse period
        if period_val:

            # Take a snippet of this pulse period
            tmp_pb_name = pb_obj.name + '_{}'.format(pb_name_idx)

            # Snippet edges
            start_t = period_dict['start_ar'][period_idx] * dur_quant
            if period_idx < last_period_idx:
                # pulse period in the bulk - take the snippet by the period edges
                stop_t = (period_dict['start_ar'][period_idx] +
                          period_dict['len_ar'][period_idx]) * dur_quant
            else:
                # right-most pulse period must be extended to include the reminder
                stop_t = pb_obj.dur

            tmp_pb = pb_snip(
                pb_obj=pb_obj,
                start_t=start_t,
                stop_t=stop_t,
                snip_name=tmp_pb_name,
                use_centers=True
            )

            # Add the snippet to the PulseBlock list
            final_snip_list.append(tmp_pb)
            # Add corresponding entry to the sequence table:
            #   output tmp_pb_name once
            final_seq_list.append(
                (tmp_pb_name, 1)
            )
            # Increment name index for the next PulseBlock
            pb_name_idx += 1

        # Wait period
        else:
            # Only add an entry to the sequence table:
            #   output 'wait waveform n times'
            final_seq_list.append(
                (wait_pb_name, period_dict['len_ar'][period_idx])
            )

    return {
        'seq_list': final_seq_list,
        'snip_list': final_snip_list
    }


def pb_expand_test(res_dict, indicate_bounds=True):
    """ Helper method to test pb_zip()

    This method takes direct output res_dict of pb_zip() call and reconstructs
    the original pulse block according to the sequence.
    To visualize boundaries between individual sequence elements, set optional
    indicate_bounds parameter to True. This will add zero-length divider pulses
    on one of the channels.

    :param res_dict: (dict) return of pb_zip() call to be reconstructed back
    into a plain waveform.
    :param indicate_bounds: (bool) if True, divider pulses will be added to one
    of the channels after each individual waveform.
    :return: (PulseBlock) reconstructed plain waveform
    """

    import pulseblock.pulse as po

    new_pb = pb.PulseBlock()

    seq_list = res_dict['seq_list']
    snip_list = res_dict['snip_list']

    for snip_idx in range(len(seq_list)):
        snip_name, rep_num = seq_list[snip_idx]

        for rep_idx in range(rep_num):

            # Find wfm_snip corresponding to snip_name
            wfm_snip = None
            for wfm_snip in snip_list:
                if wfm_snip.name == snip_name:
                    break

            new_pb.append_pb(
                pb_obj=wfm_snip
            )

            # Add a marker to indicate the block boundary
            if indicate_bounds:
                mrk_ch = list(
                    wfm_snip.dflt_dict.keys()
                )[0]

                new_pb.append(
                    p_obj=po.PTrue(
                        ch=mrk_ch,
                        dur=0
                    ),
                    cflct_er=False
                )

    return new_pb


# Technical methods


def merge_intervals(i_list):
    """ Merge overlapping intervals from a (un-ordered) list

    Each interval should be represented by a tuple
        (interval_beginning_val, interval_end_val )

    :param i_list: list of tuples - original intervals

    :return: list of tuples - list of merged intervals.
    These intervals do not overlap and are sorted in ascending order.
    """

    sorted_i_list = sorted(i_list, key=lambda i_tup: i_tup[0])
    merged_i_list = []

    for idx in range(len(sorted_i_list)):

        # In the very begining, just add the left-most interval
        # to merged_i_list
        if idx == 0:
            merged_i_list.append(sorted_i_list[0])

        else:
            # Pop-out the right-most element from merged_i_list
            # [pop-outing is necessary because the interval
            # will be changed (extended) if merging is performed]
            left_i_tup = merged_i_list.pop()

            # Check for overlap with the next interval to the right from
            # sorted_i_list
            right_i_tup = sorted_i_list[idx]

            if left_i_tup[1] >= right_i_tup[0]:
                # The two intervals overlap:
                #   merge them and append the extended interval to merged_i_list
                new_tup = tuple(
                    [
                        left_i_tup[0],
                        max(left_i_tup[1], right_i_tup[1])
                    ]
                )
                merged_i_list.append(new_tup)

            else:
                # The two intervals do not overlap:
                #   return left_i_tup to its' original place in the merged_i_list and
                #   append right_i_tup - the new right-most interval to be checked for
                #   overlap with subsequent intervals from sorted_i_list
                merged_i_list.append(left_i_tup)
                merged_i_list.append(right_i_tup)

    return merged_i_list


def run_len_encode(in_ar):
    """ Run-length encode the input array:
    finds runs of identical values and replaces them with
    {run start index, run length, run value}.

    :param in_ar: input array
    :return: (dict)
        {
            'start_ar': run start indices [np.array(np.int)]
            'len_ar': run lengths [np.array(np.int)]
            'val_ar': values within each run [np.array(dtype of in_ar)]
        }
    """

    ar_len = len(in_ar)
    in_ar = np.asarray(in_ar)

    if ar_len == 0:
        return {
            'start_ar': np.array([]),
            'len_ar': np.array([]),
            'val_ar': np.array([])
        }

    # Run starting indices
    # - pairwise comparison:
    #   check whether the next element coincides with current
    pairwise_comp = in_ar[1:] != in_ar[:-1]

    # - run start indices
    start_ar = np.concatenate((
        [0],  # the first element is always a beginning of a run
        np.where(pairwise_comp)[0] + 1  # beginnings of all subsequent runs
    ))
    # [0] is needed because np.where() returns a tuple of arrays for each dimension
    # +1 is needed because pairwise_comp contains indices of elements to the left
    # from the run boundaries, whereas starting points are the ones to the right

    # Run lengths
    len_ar = np.diff(start_ar)
    # append length of the right-most run
    len_ar = np.append(
        len_ar,
        ar_len - start_ar[-1]
    )

    # Values within each run
    val_ar = np.take(
        a=in_ar,
        indices=start_ar
    )

    return {
        'start_ar': start_ar,
        'len_ar': len_ar,
        'val_ar': val_ar
    }


def pb_snip(pb_obj, start_t, stop_t, snip_name=None, use_centers=False):
    """ PulseBlock snipping tool

    Copies all default values and all pulse objects between the snippet edges
    with preserved order and spacing between them.
    The start_t and stop_t edges are the origin and the end of the returned
    PulseBlock object.

    Note that snippet edge must not cross any pulse object (but can touch it).
    If any edge crosses any pulse object, ValueError exception is produced.

    Assumptions:
    - pulse objects are ascending-time-ordered within pb_obj.p_dict[ch_name]
    - all pulse objects have non-zero duration. The algorithm will still work
     for a zero-duration pulse, but it might be ether left out or included
     into the snippet unpredictably, if it touches the edges star_t or stop_t.

    :param pb_obj: original PulseBlock
    :param start_t: (float) left snippet edge
    :param stop_t: (float) right snippet edge
    :param snip_name: (str, optional) name for the new PulseBlock object
                      If not given, pb_obj.name + '_snip' is used.
    :param use_centers: (bool) if False, belonging of a pulse to the snippet
    is determined by edges. If True - by position of the pulse central point

    :return: new PulseBlock object -
             snippet of pb_obj between start_t and stop_t
             Exception is produced in the case of error.
    """

    # New PulseBlock to be filled and returned
    new_pb = pb.PulseBlock()
    new_pb.dur = stop_t - start_t
    if snip_name is not None:
        new_pb.name = snip_name
    else:
        new_pb.name = pb_obj.name + '_snip'

    # Copy all default pulses
    new_pb.dflt_dict = copy.deepcopy(pb_obj.dflt_dict)

    # Copy all pulses falling in between start_t and stop_t
    for ch in pb_obj.p_dict.keys():
        for pulse in pb_obj.p_dict[ch]:

            # Assumption: pulses in pb_obj.p_dict[ch] are time-ordered
            # Assumption: all pulses have dur > 0.
            #   The algorithm will also work for a pulse of dur = 0,
            #   but such pulse might be left out or included unpredictably,
            #   if it touches the window edges star_t or stop_t.

            # Two ways to determine belonging of a block to the snippet:
            #
            # 1) whether the center of a pulse falls into [start_t, stop_t]
            # interval.
            #   Pros: robust against float-comparison errors if the pulse
            #     touches snippet edge.
            #   Cons: This method does not detect the cases when snippet
            #     edges go across a pulse block.
            #
            # 2) whether both edges of a pulse fall into [start_t, stop_t]
            #   Pros: checks for snippet edges crossing the pulse
            #   Cons: may unpredictably produce exceptions if one of the pulses
            #       touches snippet edge (due to unpredictable comparison result
            #       for two nominally identical floats)

            # Method 1: use pulse center
            if use_centers:
                pulse_center = pulse.t0 + pulse.dur/2
                if pulse_center <= start_t:
                    # pulse is fully to the left from snip window
                    continue

                elif start_t <= pulse_center <= stop_t:
                    # pulse is fully inside the snip window

                    if ch not in new_pb.p_dict.keys():
                        # the first time the pulse is added to this channel
                        new_pb.p_dict[ch] = []

                    pulse_copy = copy.deepcopy(pulse)
                    pulse_copy.t0 -= start_t
                    new_pb.p_dict[ch].append(pulse_copy)

                else:
                    # This pulse and all subsequent ones lie to the right
                    # from the snip window
                    break

            # Method 2: use pulse edges
            else:
                if pulse.t0+pulse.dur <= start_t:
                    # pulse is fully to the left from snip window
                    continue

                elif start_t <= pulse.t0 and pulse.t0+pulse.dur <= stop_t:
                    # pulse is fully inside the snip window

                    if ch not in new_pb.p_dict.keys():
                        # the first time the pulse is added to this channel
                        new_pb.p_dict[ch] = []

                    pulse_copy = copy.deepcopy(pulse)
                    pulse_copy.t0 -= start_t
                    new_pb.p_dict[ch].append(pulse_copy)

                elif stop_t <= pulse.t0:
                    # This pulse and all subsequent ones lie to the right
                    # from the snip window
                    break

                # The window edge goes across one of the pulse objects.
                # Since it is not clear what to do with such a pulse,
                # one has to raise an exception.
                # Alternatively, it might be some unexpected comparison error.
                else:
                    # Determine the conflicting edge
                    if pulse.t0 < start_t < pulse.t0+pulse.dur:
                        edge_type = 'start_t'
                        edge_t = start_t

                    elif pulse.t0 < stop_t < pulse.t0+pulse.dur:
                        edge_type = 'stop_t'
                        edge_t = stop_t

                    # Something completely unexpected:
                    # if it were just a crossing with an edge,
                    # one of the above conditions should have been satisfied.
                    # Hence, this part is reached only in the case of some
                    # unexpected error.
                    #
                    # Just raise an exception and provide all the information
                    else:
                        raise ValueError(
                            'pb_snip(): condition check failed: \n'
                            '   pulse_obj = {} \n'
                            '   t0 = {} \n'
                            '   dur = {} \n'
                            '   start_t = {} \n'
                            '   stop_t = {}'
                            ''.format(
                                pulse,
                                pulse.t0,
                                pulse.dur,
                                start_t,
                                stop_t
                            )
                        )

                    raise ValueError(
                        'pb_snip(): snip edge goes across a pulse object \n'
                        '   channel = {} \n'
                        '   t0 = {} \n'
                        '   t0+dur = {} \n'
                        '   pulse object = {} \n'
                        '   conflicting edge = {} \n'
                        '   edge position = {}'
                        ''.format(
                            ch,
                            pulse.t0,
                            pulse.t0 + pulse.dur,
                            str(pulse),
                            edge_type,
                            edge_t
                        )
                    )

    return new_pb
